# This module is part of CAiMIRA. Please see the repository at
# https://gitlab.cern.ch/caimira/caimira for details of the license and terms of use.
"""
This module implements the core CAiMIRA models.

The CAiMIRA model is a flexible, object-oriented numerical model. It is designed
to allow the user to swap-out and extend its various components. One of the
major abstractions of the model is the distinction between virus concentration
(:class:`ConcentrationModel`) and virus exposure (:class:`ExposureModel`).

The concentration component is a recursive (on model time) model and therefore in order
to optimise its execution certain layers of caching are implemented. This caching
mandates that the models in this module, once instantiated, are immutable and
deterministic (i.e. running the same model twice will result in the same answer).

In order to apply stochastic / non-deterministic analyses therefore you must
introduce the randomness before constructing the models themselves; the
:mod:`caimira.monte_carlo` module is a good example of doing this - that module uses
the models defined here to allow you to construct a ConcentrationModel containing
parameters which are expressed as probability distributions. Under the hood the
``caimira.monte_carlo.ConcentrationModel`` implementation simply samples all of those
probability distributions to produce many instances of the deterministic model.

The models in this module have been designed for flexibility above performance,
particularly in the single-model case. By using the natural expressiveness of
Python we benefit from a powerful, readable and extendable implementation. A
useful feature of the implementation is that we are able to benefit from numpy
vectorisation in the case of wanting to run multiple-parameterisations of the model
at the same time. In order to benefit from this feature you must construct the models
with an array of parameter values. The values must be either scalar, length 1 arrays,
or length N arrays, where N is the number of parameterisations to run; N must be
the same for all parameters of a single model.

"""
from dataclasses import dataclass
import typing

import numpy as np
from scipy.interpolate import interp1d
import scipy.stats as sct
from scipy.optimize import minimize

from caimira.calculator.store.data_registry import DataRegistry

if not typing.TYPE_CHECKING:
    from memoization import cached
else:
    # Workaround issue https://github.com/lonelyenvoy/python-memoization/issues/18
    # by providing a no-op cache decorator when type-checking.
    cached = lambda *cached_args, **cached_kwargs: lambda function: function  # noqa

from .utils import method_cache

from .dataclass_utils import nested_replace

oneoverln2 = 1 / np.log(2)
# Define types for items supporting vectorisation. In the future this may be replaced
# by ``np.ndarray[<type>]`` once/if that syntax is supported. Note that vectorization
# implies 1d arrays: multi-dimensional arrays are not supported.
_VectorisedFloat = typing.Union[float, np.ndarray]
_VectorisedInt = typing.Union[int, np.ndarray]

Time_t = typing.TypeVar('Time_t', float, int)
BoundaryPair_t = typing.Tuple[Time_t, Time_t]
BoundarySequence_t = typing.Union[typing.Tuple[BoundaryPair_t, ...], typing.Tuple]


@dataclass(frozen=True)
class Interval:
    """
    Represents a collection of times in which a "thing" happens.

    The "thing" may be when an action is taken, such as opening a window, or
    entering a room.

    Note that all intervals are open at the start, and closed at the end. So a
    simple start, stop interval follows::

        start < t <= end

    """
    def boundaries(self) -> BoundarySequence_t:
        return ()

    def transition_times(self) -> typing.Set[float]:
        transitions = set()
        for start, end in self.boundaries():
            transitions.update([start, end])
        return transitions

    def triggered(self, time: float) -> bool:
        """Whether the given time falls inside this interval."""
        for start, end in self.boundaries():
            if start < time <= end:
                return True
        return False

@dataclass(frozen=True)
class SpecificInterval(Interval):
    #: A sequence of times (start, stop), in hours, that the infected person
    #: is present. The flattened list of times must be strictly monotonically
    #: increasing.
    present_times: BoundarySequence_t

    def boundaries(self) -> BoundarySequence_t:
        return self.present_times


@dataclass(frozen=True)
class PeriodicInterval(Interval):
    #: How often does the interval occur (minutes).
    period: float

    #: How long does the interval occur for (minutes).
    #: A value greater than :data:`period` signifies the event is permanently
    #: occurring, a value of 0 signifies that the event never happens.
    duration: float

    #: Time at which the first person (infected or exposed) arrives at the enclosed space.
    start: float = 0.0

    def boundaries(self) -> BoundarySequence_t:
        if self.period == 0 or self.duration == 0:
            return tuple()
        result = []
        for i in np.arange(self.start, 24, self.period / 60):
            # NOTE: It is important that the time type is float, not np.float, in
            # order to allow hashability (for caching).
            result.append((float(i), float(i+self.duration/60)))
        return tuple(result)


@dataclass(frozen=True)
class PiecewiseConstant:

    # TODO: Implement rather a periodic version (24-hour period), where
    # transition_times and values have the same length.

    #: transition times at which the function changes value (hours).
    transition_times: typing.Tuple[float, ...]

    #: values of the function between transitions
    values: typing.Tuple[_VectorisedFloat, ...]

    def __post_init__(self):
        if len(self.transition_times) != len(self.values)+1:
            raise ValueError("transition_times must contain one more element than values")
        if tuple(sorted(set(self.transition_times))) != self.transition_times:
            raise ValueError("transition_times must not contain duplicated elements and must be sorted")
        shapes = [np.array(v).shape for v in self.values]
        if not all(shapes[0] == shape for shape in shapes):
            raise ValueError("All values must have the same shape")

    def value(self, time) -> _VectorisedFloat:
        if time <= self.transition_times[0]:
            return self.values[0]
        elif time > self.transition_times[-1]:
            return self.values[-1]

        for t1, t2, value in zip(self.transition_times[:-1],
                                 self.transition_times[1:], self.values):
            if t1 < time <= t2:
                break
        return value

    def interval(self) -> Interval:
        # Build an Interval object
        present_times = []
        for t1, t2, value in zip(self.transition_times[:-1],
                                 self.transition_times[1:], self.values):
            if value:
                present_times.append((t1, t2))
        return SpecificInterval(present_times=tuple(present_times))

    def refine(self, refine_factor=10) -> "PiecewiseConstant":
        # Build a new PiecewiseConstant object with a refined mesh,
        # using a linear interpolation in-between the initial mesh points
        refined_times = np.linspace(self.transition_times[0], self.transition_times[-1],
                                    (len(self.transition_times)-1) * refine_factor+1)
        interpolator = interp1d(
            self.transition_times,
            np.concatenate([self.values, self.values[-1:]], axis=0),
            axis=0)
        return PiecewiseConstant(
            # NOTE: It is important that the time type is float, not np.float, in
            # order to allow hashability (for caching).
            tuple(float(time) for time in refined_times),
            tuple(interpolator(refined_times)[:-1]),
        )


@dataclass(frozen=True)
class IntPiecewiseConstant(PiecewiseConstant):

    #: values of the function between transitions
    values: typing.Tuple[int, ...]

    def value(self, time) -> _VectorisedFloat:
        if time <= self.transition_times[0] or time > self.transition_times[-1]:
            return 0

        for t1, t2, value in zip(self.transition_times[:-1],
                                 self.transition_times[1:], self.values):
            if t1 < time <= t2:
                break
        return value


@dataclass(frozen=True)
class Room:
    #: The total volume of the room
    volume: _VectorisedFloat

    #: The temperature inside the room (Kelvin).
    inside_temp: PiecewiseConstant = PiecewiseConstant((0, 24), (293,))

    #: The humidity in the room (from 0 to 1 - e.g. 0.5 is 50% humidity)
    humidity: _VectorisedFloat = 0.5

    #: The maximum occupation of the room - design limit
    capacity: typing.Optional[int] = None


@dataclass(frozen=True)
class _VentilationBase:
    """
    Represents a mechanism by which air can be exchanged (replaced/filtered)
    in a time dependent manner.

    The nature of the various air exchange schemes means that it is expected
    for subclasses of Ventilation to exist. Known subclasses include
    WindowOpening for window based air exchange, and HEPAFilter, for
    mechanical air exchange through a filter.

    """
    def transition_times(self, room: Room) -> typing.Set[float]:
        raise NotImplementedError("Subclass must implement")

    def air_exchange(self, room: Room, time: float) -> _VectorisedFloat:
        """
        Returns the rate at which air is being exchanged in the given room
        at a given time (in hours).

        Note that whilst the time is known inside this function, it may not
        be used to vary the result unless the specific time used is declared
        as part of a state change in the interval (e.g. when air_exchange == 0).

        """
        return 0.


@dataclass(frozen=True)
class Ventilation(_VentilationBase):
    #: The interval in which the ventilation is active.
    active: Interval

    def transition_times(self, room: Room) -> typing.Set[float]:
        return self.active.transition_times()


@dataclass(frozen=True)
class MultipleVentilation(_VentilationBase):
    """
    Represents a mechanism by which air can be exchanged (replaced/filtered)
    in a time dependent manner.

    Group together different sources of ventilations.

    """
    ventilations: typing.Tuple[_VentilationBase, ...]

    def transition_times(self, room: Room) -> typing.Set[float]:
        transitions = set()
        for ventilation in self.ventilations:
            transitions.update(ventilation.transition_times(room))
        return transitions

    def air_exchange(self, room: Room, time: float) -> _VectorisedFloat:
        """
        Returns the rate at which air is being exchanged in the given room
        at a given time (in hours).
        """
        return np.array([
            ventilation.air_exchange(room, time)
            for ventilation in self.ventilations
        ], dtype=object).sum(axis=0)


@dataclass(frozen=True)
class WindowOpening(Ventilation):
    #: The interval in which the window is open.
    active: Interval

    #: The temperature outside of the window (Kelvin).
    outside_temp: PiecewiseConstant

    #: The height of the window (m).
    window_height: _VectorisedFloat

    #: The length of the opening-gap when the window is open (m).
    opening_length: _VectorisedFloat

    #: The number of windows of the given dimensions.
    number_of_windows: int = 1

    #: Minimum difference between inside and outside temperature (K).
    min_deltaT: float = 0.1

    @property
    def discharge_coefficient(self) -> _VectorisedFloat:
        """
        Discharge coefficient (or cd_b): what portion effective area is
        used to exchange air (0 <= discharge_coefficient <= 1).
        To be implemented in subclasses.
        """
        raise NotImplementedError("Unknown discharge coefficient")

    def transition_times(self, room: Room) -> typing.Set[float]:
        transitions = super().transition_times(room)
        transitions.update(room.inside_temp.transition_times)
        transitions.update(self.outside_temp.transition_times)
        return transitions

    def air_exchange(self, room: Room, time: float) -> _VectorisedFloat:
        # If the window is shut, no air is being exchanged.
        if not self.active.triggered(time):
            return 0.

        # Reminder, no dependence on time in the resulting calculation.
        inside_temp: _VectorisedFloat = room.inside_temp.value(time)
        outside_temp: _VectorisedFloat = self.outside_temp.value(time)

        # The inside_temperature is forced to be always at least min_deltaT degree
        # warmer than the outside_temperature. Further research needed to
        # handle the buoyancy driven ventilation when the temperature gradient
        # is inverted.
        inside_temp = np.maximum(inside_temp, outside_temp + self.min_deltaT)  # type: ignore
        temp_gradient = (inside_temp - outside_temp) / outside_temp
        root = np.sqrt(9.81 * self.window_height * temp_gradient)
        window_area = self.window_height * self.opening_length * self.number_of_windows
        return (3600 / (3 * room.volume)) * self.discharge_coefficient * window_area * root


@dataclass(frozen=True)
class SlidingWindow(WindowOpening):
    """
    Sliding window, or side-hung window (with the hinge perpendicular to
    the horizontal plane).
    """
    data_registry: DataRegistry = DataRegistry()

    @property
    def discharge_coefficient(self) -> _VectorisedFloat:
        """
        Average measured value of discharge coefficient for sliding or
        side-hung windows.
        """
        return self.data_registry.ventilation['natural']['discharge_factor']['sliding'] # type: ignore


@dataclass(frozen=True)
class HingedWindow(WindowOpening):
    """
    Top-hung or bottom-hung hinged window (with the hinge parallel to
    horizontal plane).
    """
    #: Window width (m).
    window_width: _VectorisedFloat = 0.0

    def __post_init__(self):
        if self.window_width is float(0.0):
            raise ValueError('window_width must be set')

    @property
    def discharge_coefficient(self) -> _VectorisedFloat:
        """
        Simple model to compute discharge coefficient for top or bottom
        hung hinged windows, in the absence of empirical test results
        from manufacturers.
        From an excel spreadsheet calculator (Richard Daniels, Crawford
        Wright, Benjamin Jones - 2018) from the UK government -
        see Section 8.3 of BB101 and Section 11.3 of
        ESFA Output Specification Annex 2F on Ventilation opening areas.
        """
        window_ratio = np.array(self.window_width / self.window_height)
        coefs = np.empty(window_ratio.shape + (2, ), dtype=np.float64)

        coefs[window_ratio < 0.5] = (0.06, 0.612)
        coefs[np.bitwise_and(0.5 <= window_ratio, window_ratio < 1)] = (0.048, 0.589)
        coefs[np.bitwise_and(1 <= window_ratio, window_ratio < 2)] = (0.04, 0.563)
        coefs[window_ratio >= 2] = (0.038, 0.548)
        M, cd_max = coefs.T

        window_angle = 2.*np.rad2deg(np.arcsin(self.opening_length/(2.*self.window_height)))
        return cd_max*(1-np.exp(-M*window_angle))


@dataclass(frozen=True)
class HEPAFilter(Ventilation):
    #: The interval in which the HEPA filter is operating.
    active: Interval

    #: The rate at which the HEPA exchanges air (when switched on)
    # in m^3/h
    q_air_mech: _VectorisedFloat

    def air_exchange(self, room: Room, time: float) -> _VectorisedFloat:
        # If the HEPA is off, no air is being exchanged.
        if not self.active.triggered(time):
            return 0.
        # Reminder, no dependence on time in the resulting calculation.
        return self.q_air_mech / room.volume


@dataclass(frozen=True)
class HVACMechanical(Ventilation):
    #: The interval in which the mechanical ventilation (HVAC) is operating.
    active: Interval

    #: The rate at which the HVAC exchanges air (when switched on)
    # in m^3/h
    q_air_mech: _VectorisedFloat

    def air_exchange(self, room: Room, time: float) -> _VectorisedFloat:
        # If the HVAC is off, no air is being exchanged.
        if not self.active.triggered(time):
            return 0.
        # Reminder, no dependence on time in the resulting calculation.
        return self.q_air_mech / room.volume


@dataclass(frozen=True)
class AirChange(Ventilation):
    #: The interval in which the ventilation is operating.
    active: Interval

    #: The rate (in h^-1) at which the ventilation exchanges all the air
    # of the room (when switched on)
    air_exch: _VectorisedFloat

    def air_exchange(self, room: Room, time: float) -> _VectorisedFloat:
        # No dependence on the room volume.
        # If off, no air is being exchanged.
        if not self.active.triggered(time):
            return 0.
        # Reminder, no dependence on time in the resulting calculation.
        return self.air_exch


@dataclass(frozen=True)
class CustomVentilation(_VentilationBase):
    # The ventilation value for a given time
    ventilation_value: PiecewiseConstant

    def transition_times(self, room: Room) -> typing.Set[float]:
        return set(self.ventilation_value.transition_times)

    def air_exchange(self, room: Room, time: float) -> _VectorisedFloat:
        return self.ventilation_value.value(time)


@dataclass(frozen=True)
class Virus:
    #: RNA copies  / mL
    viral_load_in_sputum: _VectorisedFloat

    #: Dose to initiate infection, in RNA copies
    infectious_dose: _VectorisedFloat

    #: viable-to-RNA virus ratio as a function of the viral load
    viable_to_RNA_ratio: _VectorisedFloat

    #: Reported increase of transmissibility of a VOC
    transmissibility_factor: float

    #: Pre-populated examples of Viruses.
    types: typing.ClassVar[typing.Dict[str, "Virus"]]

    #: Number of days the infector is contagious
    infectiousness_days: int

    def halflife(self, humidity: _VectorisedFloat, inside_temp: _VectorisedFloat) -> _VectorisedFloat:
        # Biological decay (inactivation of the virus in air) - virus
        # dependent and function of humidity
        raise NotImplementedError

    def decay_constant(self, humidity: _VectorisedFloat, inside_temp: _VectorisedFloat) -> _VectorisedFloat:
        # Viral inactivation per hour (h^-1) (function of humidity and inside temperature)
        return np.log(2) / self.halflife(humidity, inside_temp)


@dataclass(frozen=True)
class SARSCoV2(Virus):
    #: Number of days the infector is contagious
    infectiousness_days: int = 14

    def halflife(self, humidity: _VectorisedFloat, inside_temp: _VectorisedFloat) -> _VectorisedFloat:
        """
        Half-life changes with humidity level. Here is implemented a simple
        piecewise constant model (for more details see A. Henriques et al,
        CERN-OPEN-2021-004, DOI: 10.17181/CERN.1GDQ.5Y75)
        """
        # Updated to use the formula from Dabish et al. with correction https://doi.org/10.1080/02786826.2020.1829536
        # with a maximum at hl = 6.43 (compensate for the negative decay values in the paper).
        # Note that humidity is in percentage and inside_temp in °C.
        # factor np.log(2) -> decay rate to half-life; factor 60 -> minutes to hours
        hl_calc = ((np.log(2)/((0.16030 + 0.04018*(((inside_temp-273.15)-20.615)/10.585)
                                       +0.02176*(((humidity*100)-45.235)/28.665)
                                       -0.14369
                                       -0.02636*((inside_temp-273.15)-20.615)/10.585)))/60)

        return np.where(hl_calc <= 0, 6.43, np.minimum(6.43, hl_calc))


# Example of Viruses only used for the Expert app and tests.
Virus.types = {
    'SARS_CoV_2': SARSCoV2(
        viral_load_in_sputum=1e9,
        # No data on coefficient for SARS-CoV-2 yet.
        # It is somewhere between 1000 or 10 SARS-CoV viruses,
        # as per https://www.dhs.gov/publication/st-master-question-list-covid-19
        # 50 comes from Buonanno et al.
        infectious_dose=50.,
        viable_to_RNA_ratio = 0.5,
        transmissibility_factor=1.0,
    ),
    'SARS_CoV_2_ALPHA': SARSCoV2(
        # Also called VOC-202012/01
        viral_load_in_sputum=1e9,
        infectious_dose=50.,
        viable_to_RNA_ratio = 0.5,
        transmissibility_factor=0.78,
    ),
    'SARS_CoV_2_BETA': SARSCoV2(
        viral_load_in_sputum=1e9,
        infectious_dose=50.,
        viable_to_RNA_ratio=0.5,
        transmissibility_factor=0.8,
        ),
    'SARS_CoV_2_GAMMA': SARSCoV2(
        viral_load_in_sputum=1e9,
        infectious_dose=50.,
        viable_to_RNA_ratio = 0.5,
        transmissibility_factor=0.72,
    ),
    'SARS_CoV_2_DELTA': SARSCoV2(
        viral_load_in_sputum=1e9,
        infectious_dose=50.,
        viable_to_RNA_ratio = 0.5,
        transmissibility_factor=0.51,
    ),
    'SARS_CoV_2_OMICRON': SARSCoV2(
        viral_load_in_sputum=1e9,
        infectious_dose=50.,
        viable_to_RNA_ratio=0.5,
        transmissibility_factor=0.2
    ),
}


@dataclass(frozen=True)
class Mask:
    #: Filtration efficiency of masks when inhaling.
    η_inhale: _VectorisedFloat

    #: Filtration efficiency of masks when exhaling.
    η_exhale: typing.Union[None, _VectorisedFloat] = None

    #: Global factor applied to filtration efficiency of masks when exhaling.
    factor_exhale: float = 1.

    #: Pre-populated examples of Masks.
    types: typing.ClassVar[typing.Dict[str, "Mask"]]

    def exhale_efficiency(self, diameter: _VectorisedFloat) -> _VectorisedFloat:
        """
        Overall exhale efficiency, including the effect of the leaks.
        See CERN-OPEN-2021-004 (doi: 10.17181/CERN.1GDQ.5Y75), and Ref.
        therein (Asadi 2020).
        Obtained from measurements of filtration efficiency and of
        the leakage through the sides.
        Diameter is in microns.
        """
        if self.η_exhale is not None:
            # When η_exhale is specified, return it directly
            return self.η_exhale

        d = np.array(diameter)
        intermediate_range1 = np.bitwise_and(0.5 <= d, d < 0.94614)
        intermediate_range2 = np.bitwise_and(0.94614 <= d, d < 3.)

        eta_out = np.empty(d.shape, dtype=np.float64)

        eta_out[d < 0.5] = 0.
        eta_out[intermediate_range1] = 0.5893 * d[intermediate_range1] + 0.1546
        eta_out[intermediate_range2] = 0.0509 * d[intermediate_range2] + 0.664
        eta_out[d >= 3.] = 0.8167

        return eta_out*self.factor_exhale

    def inhale_efficiency(self) -> _VectorisedFloat:
        """
        Overall inhale efficiency, including the effect of the leaks.
        """
        return self.η_inhale


# Example of Masks only used for the Expert app and tests.
Mask.types = {
    'No mask': Mask(0, 0),
    'Type I': Mask(
        η_inhale=0.5,  # (CERN-OPEN-2021-004)
    ),
    'FFP2': Mask(
        η_inhale=0.865,  # (94% penetration efficiency + 8% max inward leakage -> EN 149)
    ),
    'Cloth': Mask(  # https://doi.org/10.1080/02786826.2021.1890687
        η_inhale=0.225,
        η_exhale=0.35,
    ),
}


@dataclass(frozen=True)
class Particle:
    """
    Represents an aerosol particle.
    """

    #: diameter of the aerosol in microns
    diameter: typing.Union[None,_VectorisedFloat] = None

    def settling_velocity(self, evaporation_factor: float=0.3) -> _VectorisedFloat:
        """
        Settling velocity (i.e. speed of deposition on the floor due
        to gravity), for aerosols, in m/s. Diameter-dependent expression
        from https://doi.org/10.1101/2021.10.14.21264988
        When an aerosol-diameter is not given, returns
        the default value of 1.88e-4 m/s (corresponds to diameter of
        2.5 microns, i.e. geometric average of the breathing
        expiration distribution, taking evaporation into account, see
        https://doi.org/10.1101/2021.10.14.21264988)
        evaporation_factor represents the factor applied to the diameter,
        due to instantaneous evaporation of the particle in the air.
        """
        if self.diameter is None:
            return 1.88e-4
        else:
            return 1.88e-4 * (self.diameter*evaporation_factor / 2.5)**2

    def fraction_deposited(self, evaporation_factor: float=0.3) -> _VectorisedFloat:
        """
        The fraction of particles actually deposited in the respiratory
        tract (over the total number of particles). It depends on the
        particle diameter.
        From W. C. Hinds, New York, Wiley, 1999 (pp. 233 – 259).
        evaporation_factor represents the factor applied to the diameter,
        due to instantaneous evaporation of the particle in the air.
        """
        if self.diameter is None:
            # model is not evaluated for specific values of aerosol
            # diameters - we choose a single "average" deposition factor,
            # as in https://doi.org/10.1101/2021.10.14.21264988.
            fdep = 0.6
        else:
            # deposition fraction depends on aerosol particle diameter.
            d = (self.diameter * evaporation_factor)
            IFrac = 1 - 0.5 * (1 - (1 / (1 + (0.00076*(d**2.8)))))
            fdep = IFrac * (0.0587
                    + (0.911/(1 + np.exp(4.77 + 1.485 * np.log(d))))
                    + (0.943/(1 + np.exp(0.508 - 2.58 * np.log(d))))) # type: ignore
        return fdep


@dataclass(frozen=True)
class _ExpirationBase:
    """
    Represents the expiration of aerosols by a person.
    Subclasses of _ExpirationBase represent different models.
    """
    #: Pre-populated examples of Expirations.
    types: typing.ClassVar[typing.Dict[str, "_ExpirationBase"]]

    @property
    def particle(self) -> Particle:
        """
        The Particle object representing the aerosol - here the default one
        """
        return Particle()

    def aerosols(self, mask: Mask):
        """
        Total volume of aerosols expired per volume of exhaled air 
        considering the outward mask efficiency (mL/cm^3).
        """
        raise NotImplementedError("Subclass must implement")


@dataclass(frozen=True)
class Expiration(_ExpirationBase):
    """
    Model for the expiration. For a given diameter of aerosol, provides
    the aerosol volume, weighted by the mask outward efficiency when
    applicable.
    """
    #: diameter of the aerosol in microns
    diameter: _VectorisedFloat

    #: Total concentration of aerosols per unit volume of expired air
    # (in cm^-3), integrated over all aerosol diameters (corresponding
    # to c_n,i in Eq. (4) of https://doi.org/10.1101/2021.10.14.21264988)
    cn: float = 1.

    @property
    def particle(self) -> Particle:
        """
        The Particle object representing the aerosol
        """
        return Particle(diameter=self.diameter)

    @cached()
    def aerosols(self, mask: Mask):
        """
        Total volume of aerosols expired per volume 
        of exhaled air considering the outward mask 
        efficiency. Result is in mL.cm^-3.
        """
        def volume(d):
            return (np.pi * d**3) / 6.

        # Final result converted from microns^3/cm3 to mL/cm^3
        return self.cn * (volume(self.diameter) *
                (1 - mask.exhale_efficiency(self.diameter))) * 1e-12


@dataclass(frozen=True)
class MultipleExpiration(_ExpirationBase):
    """
    Represents an expiration of aerosols.
    Group together different modes of expiration, that represent
    each the main expiration mode for a certain fraction of time (given by
    the weights).
    This class can only be used with single diameters defined in each
    expiration (it cannot be used with diameter distributions).
    """
    expirations: typing.Tuple[_ExpirationBase, ...]
    weights: typing.Tuple[float, ...]

    def __post_init__(self):
        if len(self.expirations) != len(self.weights):
            raise ValueError("expirations and weigths must contain the"
                             "same number of elements")
        if not all(np.isscalar(e.diameter) for e in self.expirations):
            raise ValueError("diameters must all be scalars")

    def aerosols(self, mask: Mask):
        return np.array([
            weight * expiration.aerosols(mask) / sum(self.weights)
            for weight,expiration in zip(self.weights,self.expirations)
        ]).sum(axis=0)


# Typical expirations. The aerosol diameter given is an equivalent
# diameter, chosen in such a way that the aerosol volume is
# the same as the total aerosol volume given by the full BLO model
# (integrated between 0.1 and 30 microns)
# The correspondence with the BLO coefficients is given.
# Only used for the Expert app and tests.
_ExpirationBase.types = {
    'Breathing': Expiration(1.3844), # corresponds to B/L/O coefficients of (1, 0, 0)
    'Speaking': Expiration(5.8925),   # corresponds to B/L/O coefficients of (1, 1, 1)
    'Shouting': Expiration(10.0411), # corresponds to B/L/O coefficients of (1, 5, 5)
    'Singing': Expiration(10.0411),  # corresponds to B/L/O coefficients of (1, 5, 5)
}


@dataclass(frozen=True)
class Activity:
    #: Inhalation rate in m^3/h
    inhalation_rate: _VectorisedFloat

    #: Exhalation rate in m^3/h
    exhalation_rate: _VectorisedFloat

    #: Pre-populated examples of Activities.
    types: typing.ClassVar[typing.Dict[str, "Activity"]]


# Example of Activities only used for the Expert app and tests.
Activity.types = {
    'Seated': Activity(0.51, 0.51),
    'Standing': Activity(0.57, 0.57),
    'Light activity': Activity(1.25, 1.25),
    'Moderate activity': Activity(1.78, 1.78),
    'Heavy exercise': Activity(3.30, 3.30),
}


@dataclass(frozen=True)
class SimplePopulation:
    """
    Represents a group of people all with exactly the same behaviour and
    situation.

    """
    #: How many in the population.
    number: typing.Union[int, IntPiecewiseConstant]

    #: The times in which the people are in the room.
    presence: typing.Optional[Interval]

    #: The physical activity being carried out by the people.
    activity: Activity

    def __post_init__(self):
        if isinstance(self.number, int):
            if not isinstance(self.presence, Interval):
                raise TypeError(f'The presence argument must be an "Interval". Got {type(self.presence)}')
        else:
            if self.presence is not None:
                raise TypeError(f'The presence argument must be None for a IntPiecewiseConstant number')

    def presence_interval(self):
        if isinstance(self.presence, Interval):
            return self.presence
        elif isinstance(self.number, IntPiecewiseConstant):
            return self.number.interval()

    def person_present(self, time: float):
        # Allow back-compatibility
        if isinstance(self.number, int) and isinstance(self.presence, Interval):
            return self.presence.triggered(time)
        elif isinstance(self.number, IntPiecewiseConstant):
            return self.number.value(time) != 0

    def people_present(self, time: float):
        # Allow back-compatibility
        if isinstance(self.number, int):
            return self.number * self.person_present(time)
        else:
            return int(self.number.value(time))


@dataclass(frozen=True)
class Population(SimplePopulation):
    """
    Represents a group of people all with exactly the same behaviour and
    situation, considering the usage of mask and a certain host immunity.

    """
    #: The kind of mask being worn by the people.
    mask: Mask

    #: The ratio of virions that are inactivated by the person's immunity.
    # This parameter considers the potential antibodies in the person,
    # which might render inactive some RNA copies (virions).
    host_immunity: float


@dataclass(frozen=True)
class _PopulationWithVirus(Population):
    data_registry: DataRegistry

    #: The virus with which the population is infected.
    virus: Virus

    @method_cache
    def fraction_of_infectious_virus(self) -> _VectorisedFloat:
        """
        The fraction of infectious virus.

        """
        return 1

    def aerosols(self):
        """
        Total volume of aerosols expired per volume of exhaled air (mL/cm^3).
        """
        raise NotImplementedError("Subclass must implement")

    def emission_rate_per_aerosol_per_person_when_present(self) -> _VectorisedFloat:
        """
        The emission rate of infectious respiratory particles (IRP) in the expired air per 
        mL of respiratory fluid, if the infected population is present, in (virions.cm^3)/(mL.h).
        This method returns only the diameter-independent variables within the emission rate.
        It should not be a function of time.
        """
        raise NotImplementedError("Subclass must implement")

    @method_cache
    def emission_rate_per_person_when_present(self) -> _VectorisedFloat:
        """
        The emission rate if the infected population is present, per person
        (in virions/h).
        """
        return (self.emission_rate_per_aerosol_per_person_when_present() *
                self.aerosols())

    def emission_rate(self, time: float) -> _VectorisedFloat:
        """
        The emission rate of the population vs time.
        """
        # Note: The original model avoids time dependence on the emission rate
        # at the cost of implementing a piecewise (on time) concentration function.

        if not self.person_present(time):
            return 0.

        # Note: It is essential that the value of the emission rate is not
        # itself a function of time. Any change in rate must be accompanied
        # with a declaration of state change time, as is the case for things
        # like Ventilation.
        return self.emission_rate_per_person_when_present() * self.people_present(time)

    @property
    def particle(self) -> Particle:
        """
        The Particle object representing the aerosol expired by the
        population - here we take the default Particle object
        """
        return Particle()


@dataclass(frozen=True)
class EmittingPopulation(_PopulationWithVirus):
    #: The emission rate of a single individual, in virions / h.
    known_individual_emission_rate: float

    def aerosols(self):
        """
        Total volume of aerosols expired per volume of exhaled air (mL/cm^3).
        Here arbitrarily set to 1 as the full emission rate is known.
        """
        return 1.

    @method_cache
    def emission_rate_per_aerosol_per_person_when_present(self) -> _VectorisedFloat:
        """
        The emission rate of infectious respiratory particles (IRP) in the expired air per 
        mL of respiratory fluid, if the infected population is present, in (virions.cm^3)/(mL.h).
        This method returns only the diameter-independent variables within the emission rate.
        It should not be a function of time.
        """
        return self.known_individual_emission_rate
    

@dataclass(frozen=True)
class InfectedPopulation(_PopulationWithVirus):
    #: The type of expiration that is being emitted whilst doing the activity.
    expiration: _ExpirationBase

    @method_cache
    def fraction_of_infectious_virus(self) -> _VectorisedFloat:
        """
        The fraction of infectious virus.
        """
        return self.virus.viable_to_RNA_ratio * (1 - self.host_immunity)

    def aerosols(self):
        """
        Total volume of aerosols expired per volume of exhaled air (mL/cm^3).
        """
        return self.expiration.aerosols(self.mask)

    @method_cache
    def emission_rate_per_aerosol_per_person_when_present(self) -> _VectorisedFloat:
        """
        The emission rate of infectious respiratory particles (IRP) in the expired air per 
        mL of respiratory fluid, if the infected population is present, in (virions.cm^3)/(mL.h).
        This method returns only the diameter-independent variables within the emission rate.
        It should not be a function of time.
        """
        # Conversion factor explanation:
        # The exhalation rate is in m^3/h, therefore the 1e6 conversion factor
        # is to convert m^3/h into cm^3/h to return (virions.cm^3)/(mL.h),
        # so that we can then multiply by aerosols (mL/cm^3).
        ER = (self.virus.viral_load_in_sputum *
              self.activity.exhalation_rate *
              self.fraction_of_infectious_virus() *
              10**6)
        return ER

    @property
    def particle(self) -> Particle:
        """
        The Particle object representing the aerosol - here the default one
        """
        return self.expiration.particle


@dataclass(frozen=True)
class Cases:
    """
    The geographical data to calculate the probability of having at least 1
    new infection in a probabilistic exposure.
    """
    #: Geographic location population
    geographic_population: int = 0

    #: Geographic location new cases
    geographic_cases: int = 0

    #: Number of new cases confidence level
    ascertainment_bias: int = 0

    def probability_random_individual(self, virus: Virus) -> _VectorisedFloat:
        """Probability that a randomly selected individual in a focal population is infected."""
        return self.geographic_cases*virus.infectiousness_days*self.ascertainment_bias/self.geographic_population

    def probability_meet_infected_person(self, virus: Virus, n_infected: int, event_population: int) -> _VectorisedFloat:
        """
        Probability to meet n_infected persons in an event.
        From https://doi.org/10.1038/s41562-020-01000-9.
        """
        return sct.binom.pmf(n_infected, event_population, self.probability_random_individual(virus))


@dataclass(frozen=True)
class _ConcentrationModelBase:
    """
    A generic superclass that contains the methods to calculate the
    concentration (e.g. viral concentration or CO2 concentration).
    """
    data_registry: DataRegistry
    room: Room
    ventilation: _VentilationBase

    @property
    def population(self) -> SimplePopulation:
        """
        Population in the room (the emitters of what we compute the
        concentration of)
        """
        raise NotImplementedError("Subclass must implement")

    def removal_rate(self, time: float) -> _VectorisedFloat:
        """
        Remove rate of the species considered, in h^-1
        """
        raise NotImplementedError("Subclass must implement")

    def min_background_concentration(self) -> _VectorisedFloat:
        """
        Minimum background concentration in the room for a given scenario
        (in the same unit as the concentration). Its the value towards which
        the concentration will decay to.
        """
        return self.data_registry.concentration_model['virus_concentration_model']['min_background_concentration'] # type: ignore

    def normalization_factor(self) -> _VectorisedFloat:
        """
        Normalization factor (in the same unit as the concentration).
        This factor is applied to the normalized concentration only
        at the very end.
        """
        raise NotImplementedError("Subclass must implement")

    @method_cache
    def _normed_concentration_limit(self, time: float) -> _VectorisedFloat:
        """
        Provides a constant that represents the theoretical asymptotic
        value reached by the concentration when time goes to infinity,
        if all parameters were to stay time-independent.
        This is normalized by the normalization factor, the latter acting as a
        multiplicative constant factor for the concentration model that
        can be put back in front of the concentration after the time
        dependence has been solved for.
        """
        V = self.room.volume
        RR = self.removal_rate(time)

        if isinstance(RR, np.ndarray):
            invRR = np.empty(RR.shape, dtype=np.float64)
            invRR[RR == 0.] = np.nan
            invRR[RR != 0.] = 1. / RR[RR != 0.]
        else:
            invRR = np.nan if RR == 0. else 1. / RR # type: ignore

        return (self.population.people_present(time) * invRR / V +
                self.min_background_concentration()/self.normalization_factor())

    @method_cache
    def state_change_times(self) -> typing.List[float]:
        """
        All time dependent entities on this model must provide information about
        the times at which their state changes.
        """
        state_change_times = {0.}
        state_change_times.update(self.population.presence_interval().transition_times())
        state_change_times.update(self.ventilation.transition_times(self.room))
        return sorted(state_change_times)

    @method_cache
    def _first_presence_time(self) -> float:
        """
        First presence time. Before that, the concentration is zero.
        """
        return self.population.presence_interval().boundaries()[0][0]

    def last_state_change(self, time: float) -> float:
        """
        Find the most recent/previous state change.

        Find the nearest time less than the given one. If there is a state
        change exactly at ``time`` the previous state change is returned
        (except at ``time == 0``).
        """
        times = self.state_change_times()
        t_index: int = np.searchsorted(times, time)  # type: ignore
        # Search sorted gives us the index to insert the given time. Instead we
        # want to get the index of the most recent time, so reduce the index by
        # one unless we are already at 0.
        t_index = max([t_index - 1, 0])
        return times[t_index]

    def _next_state_change(self, time: float) -> float:
        """
        Find the nearest future state change.
        """
        for change_time in self.state_change_times():
            if change_time >= time:
                return change_time
        raise ValueError(
            f"The requested time ({time}) is greater than last available "
            f"state change time ({change_time})"
        )

    @method_cache
    def _normed_concentration_cached(self, time: float) -> _VectorisedFloat:
        """
        A cached version of the _normed_concentration method. Use this
        method if you expect that there may be multiple concentration
        calculations for the same time (e.g. at state change times).
        """
        return self._normed_concentration(time)

    def _normed_concentration(self, time: float) -> _VectorisedFloat:
        """
        Concentration as a function of time, and normalized by
        normalization_factor.
        The formulas used here assume that all parameters (ventilation,
        emission rate) are constant between two state changes - only
        the value of these parameters at the next state change, are used.

        Note that time is not vectorised. You can only pass a single float
        to this method.
        """
        # The model always starts at t=0, but we avoid running concentration calculations
        # before the first presence as an optimisation.
        if time <= self._first_presence_time():
            return self.min_background_concentration()/self.normalization_factor()

        next_state_change_time = self._next_state_change(time)
        RR = self.removal_rate(next_state_change_time)

        t_last_state_change = self.last_state_change(time)
        conc_at_last_state_change = self._normed_concentration_cached(t_last_state_change)
        delta_time = time - t_last_state_change

        fac = np.exp(-RR * delta_time)
        if isinstance(RR, np.ndarray):
            curr_conc_state = np.empty(RR.shape, dtype=np.float64)
            curr_conc_state[RR == 0.] = delta_time * self.population.people_present(time) / (
                self.room.volume[RR == 0.] if isinstance(self.room.volume,np.ndarray) else self.room.volume)
            curr_conc_state[RR != 0.] = self._normed_concentration_limit(next_state_change_time)[RR != 0.] * (1 - fac[RR != 0.])
        else:
            if RR == 0.:
                curr_conc_state = delta_time * self.population.people_present(time) / self.room.volume
            else:
                curr_conc_state = self._normed_concentration_limit(next_state_change_time) * (1 - fac)

        return curr_conc_state + conc_at_last_state_change * fac

    def concentration(self, time: float) -> _VectorisedFloat:
        """
        Total concentration as a function of time. The normalization
        factor has been put back.

        Note that time is not vectorised. You can only pass a single float
        to this method.
        """
        return (self._normed_concentration_cached(time) *
                self.normalization_factor())

    @method_cache
    def normed_integrated_concentration(self, start: float, stop: float) -> _VectorisedFloat:
        """
        Get the integrated concentration between the times start and stop,
        normalized by normalization_factor.
        """
        if stop <= self._first_presence_time():
            return (stop - start)*self.min_background_concentration()/self.normalization_factor()
        state_change_times = self.state_change_times()
        req_start, req_stop = start, stop
        total_normed_concentration = 0.
        for interval_start, interval_stop in zip(state_change_times[:-1], state_change_times[1:]):
            if req_start > interval_stop or req_stop < interval_start:
                continue
            # Clip the current interval to the requested range.
            start = max([interval_start, req_start])
            stop = min([interval_stop, req_stop])

            conc_start = self._normed_concentration_cached(start)

            next_conc_state = self._next_state_change(stop)
            conc_limit = self._normed_concentration_limit(next_conc_state)
            RR = self.removal_rate(next_conc_state)
            delta_time = stop - start
            total_normed_concentration += (
                conc_limit * delta_time +
                (conc_limit - conc_start) * (np.exp(-RR*delta_time)-1) / RR
            )
        return total_normed_concentration

    def integrated_concentration(self, start: float, stop: float) -> _VectorisedFloat:
        """
        Get the integrated concentration of viruses in the air between the times start and stop.
        """
        return (self.normed_integrated_concentration(start, stop) *
                self.normalization_factor())


@dataclass(frozen=True)
class ConcentrationModel(_ConcentrationModelBase):
    """
    Class used for the computation of the long-range virus concentration.
    """
    #: Infected population in the room, emitting virions
    infected: InfectedPopulation

    #: evaporation factor: the particles' diameter is multiplied by this
    # factor as soon as they are in the air (but AFTER going out of the,
    # mask, if any).
    evaporation_factor: float

    def __post_init__(self):
        if self.evaporation_factor is None:
            self.evaporation_factor = self.data_registry.expiration_particle['particle']['evaporation_factor']

    @property
    def population(self) -> InfectedPopulation:
        return self.infected

    @property
    def virus(self) -> Virus:
        return self.infected.virus

    def normalization_factor(self) -> _VectorisedFloat:
        # we normalize by the emission rate
        return self.infected.emission_rate_per_person_when_present()

    def removal_rate(self, time: float) -> _VectorisedFloat:
        # Equilibrium velocity of particle motion toward the floor
        vg = self.infected.particle.settling_velocity(self.evaporation_factor)
        # Height of the emission source to the floor - i.e. mouth/nose (m)
        h = 1.5
        # Deposition rate (h^-1)
        k = (vg * 3600) / h
        return (
            k + self.virus.decay_constant(self.room.humidity, self.room.inside_temp.value(time))
            + self.ventilation.air_exchange(self.room, time)
        )

    def infectious_virus_removal_rate(self, time: float) -> _VectorisedFloat:
        # defined for back-compatibility purposes
        return self.removal_rate(time)


@dataclass(frozen=True)
class CO2ConcentrationModel(_ConcentrationModelBase):
    """
    Class used for the computation of the CO2 concentration.
    """
    #: Population in the room emitting CO2
    CO2_emitters: SimplePopulation

    #: CO2 concentration in the atmosphere (in ppm)
    @property
    def CO2_atmosphere_concentration(self) -> float:
        return self.data_registry.concentration_model['CO2_concentration_model']['CO2_atmosphere_concentration'] # type: ignore

    #: CO2 fraction in the exhaled air
    @property
    def CO2_fraction_exhaled(self) -> float:
        return self.data_registry.concentration_model['CO2_concentration_model']['CO2_fraction_exhaled'] # type: ignore

    @property
    def population(self) -> SimplePopulation:
        return self.CO2_emitters

    def removal_rate(self, time: float) -> _VectorisedFloat:
        return self.ventilation.air_exchange(self.room, time)

    def min_background_concentration(self) -> _VectorisedFloat:
        """
        Background CO2 concentration in the atmosphere (in ppm)
        """
        return self.CO2_atmosphere_concentration

    def normalization_factor(self) -> _VectorisedFloat:
        # normalization by the CO2 exhaled per person.
        # CO2 concentration given in ppm, hence the 1e6 factor.
        return (1e6*self.population.activity.exhalation_rate
                *self.CO2_fraction_exhaled)


@dataclass(frozen=True)
class ShortRangeModel:
    '''
    Based on the two-stage (jet/puff) expiratory jet model by
    Jia et al (2022) - https://doi.org/10.1016/j.buildenv.2022.109166
    '''
    data_registry: DataRegistry

    #: Expiration type
    expiration: _ExpirationBase

    #: Activity type
    activity: Activity

    #: Short-range expiration and respective presence
    presence: SpecificInterval

    #: Interpersonal distances
    distance: _VectorisedFloat

    def dilution_factor(self) -> _VectorisedFloat:
        '''
        The dilution factor for the respective expiratory activity type.
        '''
        _dilution_factor = self.data_registry.short_range_model['dilution_factor'] 
        # Average mouth opening diameter (m)
        mouth_diameter: float = _dilution_factor['mouth_diameter'] # type: ignore

        # Breathing rate, from m3/h to m3/s
        BR = np.array(self.activity.exhalation_rate/3600.)

        # Exhalation coefficient. Ratio between the duration of a breathing cycle and the duration of
        # the exhalation.
        φ: float = _dilution_factor['exhalation_coefficient'] # type: ignore

        # Exhalation airflow, as per Jia et al. (2022)
        Q_exh: _VectorisedFloat = φ * BR

        # Area of the mouth assuming a perfect circle (m2)
        Am = np.pi*(mouth_diameter**2)/4

        # Initial velocity of the exhalation airflow (m/s)
        u0 = np.array(Q_exh/Am)

        # Duration of one breathing cycle
        breathing_cicle: float = _dilution_factor['breathing_cycle'] # type: ignore

        # Duration of the expiration period(s)
        tstar: float = breathing_cicle / 2

        # Streamwise and radial penetration coefficients
        _df_pc = _dilution_factor['penetration_coefficients'] # type: ignore
        𝛽r1: float = _df_pc['𝛽r1'] # type: ignore
        𝛽r2: float = _df_pc['𝛽r2'] # type: ignore
        𝛽x1: float = _df_pc['𝛽x1'] # type: ignore

        # Parameters in the jet-like stage
        # Position of virtual origin
        x0 = mouth_diameter/2/𝛽r1
        # Time of virtual origin
        t0 = (np.sqrt(np.pi)*(mouth_diameter**3))/(8*(𝛽r1**2)*(𝛽x1**2)*Q_exh)
        # The transition point, m
        xstar = np.array(𝛽x1*(Q_exh*u0)**0.25*(tstar + t0)**0.5 - x0)
        # Dilution factor at the transition point xstar
        Sxstar = np.array(2*𝛽r1*(xstar+x0)/mouth_diameter)

        distances = np.array(self.distance)
        factors = np.empty(distances.shape, dtype=np.float64)
        factors[distances < xstar] = 2*𝛽r1*(distances[distances < xstar]
                                        + x0)/mouth_diameter
        factors[distances >= xstar] = Sxstar[distances >= xstar]*(1 +
            𝛽r2*(distances[distances >= xstar] -
            xstar[distances >= xstar])/𝛽r1/(xstar[distances >= xstar]
            + x0))**3
        return factors
    
    def _normed_jet_origin_concentration(self) -> _VectorisedFloat:
        """
        The initial jet concentration at the source origin (mouth/nose), normalized by
        normalization_factor in the ShortRange class (corresponding to the diameter-independent
        variables). Results in mL.cm^-3.
        """
        # The short range origin concentration does not consider the mask contribution.
        return self.expiration.aerosols(mask=Mask.types['No mask'])

    def _long_range_normed_concentration(self, concentration_model: ConcentrationModel, time: float) -> _VectorisedFloat:
        """
        Virus long-range exposure concentration normalized by normalization_factor in the 
        ShortRange class, as function of time. Results in mL.cm^-3.
        """
        return (concentration_model.concentration(time) / self.normalization_factor(concentration_model.infected))

    def _normed_concentration(self, concentration_model: ConcentrationModel, time: float) -> _VectorisedFloat:
        """
        Virus short-range exposure concentration, as a function of time.

        If the given time falls within a short-range interval it returns the
        short-range concentration normalized by normalization_factor in the
        Short-range class. Otherwise it returns 0. Results in mL.cm^-3.
        """
        start, stop = self.presence.boundaries()[0]
        # Verifies if the given time falls within a short-range interaction
        if start <= time <= stop:
            dilution = self.dilution_factor()
            # Jet origin concentration normalized by the emission rate (except the BR)
            normed_jet_origin_concentration = self._normed_jet_origin_concentration()
            # Long-range concentration normalized by the emission rate (except the BR)
            long_range_normed_concentration = self._long_range_normed_concentration(concentration_model, time)

            # The long-range concentration values are then approximated using interpolation:
            # The set of points where we want the interpolated values are the short-range particle diameters (given the current expiration);
            # The set of points with a known value are the long-range particle diameters (given the initial expiration);
            # The set of known values are the long-range concentration values normalized by the viral load.
            long_range_normed_concentration_interpolated=np.interp(np.array(self.expiration.particle.diameter),
                                np.array(concentration_model.infected.particle.diameter), long_range_normed_concentration)

            # Short-range concentration formula. The long-range concentration is added in the concentration method (ExposureModel).
            # based on continuum model proposed by Jia et al (2022) - https://doi.org/10.1016/j.buildenv.2022.109166
            return ((1/dilution)*(normed_jet_origin_concentration - long_range_normed_concentration_interpolated))
        return 0.
    
    def normalization_factor(self, infected: InfectedPopulation) -> _VectorisedFloat:
        """
        The normalization factor applied to the short-range results. It refers to the emission
        rate per aerosol without accounting for the exhalation rate (viral load and f_inf).
        Result in (virions.cm^3)/(mL.m^3).
        """
        # Re-use the emission rate method divided by the BR contribution. 
        return infected.emission_rate_per_aerosol_per_person_when_present() / infected.activity.exhalation_rate
    
    def jet_origin_concentration(self, infected: InfectedPopulation) -> _VectorisedFloat:
        """
        The initial jet concentration at the source origin (mouth/nose).
        Returns the full result with the diameter dependent and independent variables, in virions/m^3.
        """
        return self._normed_jet_origin_concentration() * self.normalization_factor(infected)
    
    def short_range_concentration(self, concentration_model: ConcentrationModel, time: float) -> _VectorisedFloat:
        """
        Virus short-range exposure concentration, as a function of time.
        Factor of normalization applied back here. Results in virions/m^3.
        """
        return (self._normed_concentration(concentration_model, time) * 
                self.normalization_factor(concentration_model.infected))

    @method_cache
    def _normed_short_range_concentration_cached(self, concentration_model: ConcentrationModel, time: float) -> _VectorisedFloat:
        # A cached version of the _normed_concentration method. Use this
        # method if you expect that there may be multiple short-range concentration
        # calculations for the same time (e.g. at state change times).
        return self._normed_concentration(concentration_model, time)

    @method_cache
    def extract_between_bounds(self, time1: float, time2: float) -> typing.Union[None, typing.Tuple[float,float]]:
        """
        Extract the bounds of the interval resulting from the
        intersection of [time1, time2] and the presence interval.
        If [time1, time2] has nothing common to the presence interval,
        we return (0, 0).
        Raise an error if time1 and time2 are not in ascending order.
        """
        if time1>time2:
            raise ValueError("time1 must be less or equal to time2")

        start, stop = self.presence.boundaries()[0]
        if (stop < time1) or (start > time2):
            return (0, 0)
        elif start <= time1 and time2<= stop:
            return time1, time2
        elif start <= time1 and stop < time2:
            return time1, stop
        elif time1 < start and time2 <= stop:
            return start, time2
        elif time1 <= start and stop < time2:
            return start, stop

    def _normed_jet_exposure_between_bounds(self,
                    time1: float, time2: float):
        """
        Get the part of the integrated short-range concentration of
        viruses in the air, between the times start and stop, coming
        from the jet concentration, normalized by normalization_factor, 
        and without dilution.
        """
        start, stop = self.extract_between_bounds(time1, time2)
        # Note the conversion factor mL.cm^-3 -> mL.m^-3
        jet_origin = self._normed_jet_origin_concentration() * 10**6
        return jet_origin * (stop - start)

    def _normed_interpolated_longrange_exposure_between_bounds(
                    self, concentration_model: ConcentrationModel,
                    time1: float, time2: float):
        """
        Get the part of the integrated short-range concentration due
        to the background concentration, normalized by normalization_factor 
        together with breathing rate, and without dilution.
        One needs to interpolate the integrated long-range concentration
        for the particle diameters defined here.
        """
        start, stop = self.extract_between_bounds(time1, time2)
        if stop<=start:
            return 0.

        # Note that for the correct integration one needs to isolate those parameters
        # that are diameter-dependent from those that are diameter independent.
        # Therefore, the diameter-independent parameters (viral load, f_ind and BR)
        # are removed for the interpolation, and added back once the integration over
        # the new aerosol diameters (done with the mean) is completed.
        normed_int_concentration = (
            concentration_model.integrated_concentration(start, stop)
                /concentration_model.virus.viral_load_in_sputum
                /concentration_model.infected.fraction_of_infectious_virus()
                /concentration_model.infected.activity.exhalation_rate
                )
        normed_int_concentration_interpolated = np.interp(
                np.array(self.expiration.particle.diameter),
                np.array(concentration_model.infected.particle.diameter),
                normed_int_concentration
                )
        return normed_int_concentration_interpolated


@dataclass(frozen=True)
class CO2DataModel:
    '''
    The CO2DataModel class models CO2 data based on room volume and capacity, 
    ventilation transition times, and people presence.
    It uses optimization techniques to fit the model's parameters and estimate the
    exhalation rate and ventilation values that best match the measured CO2 concentrations.
    '''
    data_registry: DataRegistry
    room: Room
    occupancy: IntPiecewiseConstant
    ventilation_transition_times: typing.Tuple[float, ...]
    times: typing.Sequence[float]
    CO2_concentrations: typing.Sequence[float]

    def CO2_concentration_model(self, 
                                exhalation_rate: float, 
                                ventilation_values: typing.Tuple[float, ...]) -> CO2ConcentrationModel:
        return CO2ConcentrationModel(
            data_registry=self.data_registry,
            room=Room(volume=self.room.volume),
            ventilation=CustomVentilation(PiecewiseConstant(
                self.ventilation_transition_times, ventilation_values)),
            CO2_emitters=SimplePopulation(
                number=self.occupancy,
                presence=None,
                activity=Activity(
                    exhalation_rate=exhalation_rate, inhalation_rate=exhalation_rate),
            )
        )

    def CO2_concentrations_from_params(self, CO2_concentration_model: CO2ConcentrationModel) -> typing.List[_VectorisedFloat]:
        # Calculate the predictive CO2 concentration
        return [CO2_concentration_model.concentration(time) for time in self.times]

    def CO2_fit_params(self) -> typing.Dict:
        if len(self.times) != len(self.CO2_concentrations):
            raise ValueError('times and CO2_concentrations must have same length.')

        if len(self.times) < 2:
            raise ValueError(
                'times and CO2_concentrations must contain at last two elements')

        def fun(x):
            '''
            The objective function to be minimized, where x is an argument
            containing the initial guess for the breathing rate (exhalation_rate) 
            and ventilation values (ventilation_values).
            '''
            exhalation_rate = x[0]
            ventilation_values = tuple(x[1:])
            CO2_concentration_model = self.CO2_concentration_model(
                exhalation_rate=exhalation_rate,
                ventilation_values=ventilation_values
            )
            the_concentrations = self.CO2_concentrations_from_params(CO2_concentration_model)
            return np.sqrt(np.sum((np.array(self.CO2_concentrations) -
                                   np.array(the_concentrations))**2))
        # The goal is to minimize the difference between the two different curves (known concentrations vs. predicted concentrations)
        res_dict = minimize(fun=fun, x0=np.ones(len(self.ventilation_transition_times)), method='powell',
                            bounds=[(0, None) for _ in range(len(self.ventilation_transition_times))],
                            options={'xtol': 1e-3})

        # Final prediction
        exhalation_rate = res_dict['x'][0]
        ventilation_values = res_dict['x'][1:] # In ACH

        # Final CO2ConcentrationModel with obtained prediction
        the_CO2_concentration_model = self.CO2_concentration_model(
            exhalation_rate=exhalation_rate, 
            ventilation_values=ventilation_values
        )
        the_predictive_CO2 = self.CO2_concentrations_from_params(the_CO2_concentration_model)

        # Ventilation in L/s
        flow_rates_l_s = [vent / 3600 * self.room.volume * 1000 for vent in ventilation_values] # 1m^3 = 1000L
        
        # Ventilation in L/s/person
        flow_rates_l_s_p = [flow_rate / self.room.capacity for flow_rate in flow_rates_l_s] if self.room.capacity else None
        
        return {
            "exhalation_rate": exhalation_rate, 
            "ventilation_values": list(ventilation_values),
            "room_capacity": self.room.capacity,
            "ventilation_ls_values": flow_rates_l_s,
            "ventilation_lsp_values": flow_rates_l_s_p,
            'predictive_CO2': list(the_predictive_CO2)
        }


@dataclass(frozen=True)
class ExposureModel:
    """
    Represents the exposure to a concentration of
    infectious respiratory particles (IRP) in the air.
    """
    data_registry: DataRegistry

    #: The virus concentration model which this exposure model should consider.
    concentration_model: ConcentrationModel

    #: The list of short-range models which this exposure model should consider.
    short_range: typing.Tuple[ShortRangeModel, ...]

    #: The population of non-infected people to be used in the model.
    exposed: Population

    #: Geographical data
    geographical_data: Cases

    #: Total people with short-range interactions
    exposed_to_short_range: int = 0

    #: The number of times the exposure event is repeated (default 1).
    @property
    def repeats(self) -> int:
        return 1

    def __post_init__(self):
        """
        When diameters are sampled (given as an array),
        the Monte-Carlo integration over the diameters
        assumes that all the parameters within the IVRR,
        apart from the settling velocity, are NOT arrays.
        In other words, the air exchange rate from the
        ventilation, and the virus decay constant, must
        not be given as arrays.
        """
        c_model = self.concentration_model
        # Check if the diameter is vectorised.
        if (isinstance(c_model.infected, InfectedPopulation) and not np.isscalar(c_model.infected.expiration.diameter)
            # Check if the diameter-independent elements of the infectious_virus_removal_rate method are vectorised.
            and not (
                all(np.isscalar(c_model.virus.decay_constant(c_model.room.humidity, c_model.room.inside_temp.value(time)) +
                c_model.ventilation.air_exchange(c_model.room, time)) for time in c_model.state_change_times()))):
            raise ValueError("If the diameter is an array, none of the ventilation parameters "
                             "or virus decay constant can be arrays at the same time.")

    @method_cache
    def population_state_change_times(self) -> typing.List[float]:
        """
        All time dependent population entities on this model must provide information
        about the times at which their state changes.
        """
        state_change_times = set(self.concentration_model.infected.presence_interval().transition_times())
        state_change_times.update(self.exposed.presence_interval().transition_times())

        return sorted(state_change_times)

    def long_range_fraction_deposited(self) -> _VectorisedFloat:
        """
        The fraction of particles actually deposited in the respiratory
        tract (over the total number of particles). It depends on the
        particle diameter.
        """
        return self.concentration_model.infected.particle.fraction_deposited(
                    self.concentration_model.evaporation_factor)

    def _long_range_normed_exposure_between_bounds(self, time1: float, time2: float) -> _VectorisedFloat:
        """
        The number of virions per meter^3 between any two times, normalized
        by the emission rate of the infected population
        """
        exposure = 0.
        for start, stop in self.exposed.presence_interval().boundaries():
            if stop < time1:
                continue
            elif start > time2:
                break
            elif start <= time1 and time2<= stop:
                exposure += self.concentration_model.normed_integrated_concentration(time1, time2)
            elif start <= time1 and stop < time2:
                exposure += self.concentration_model.normed_integrated_concentration(time1, stop)
            elif time1 < start and time2 <= stop:
                exposure += self.concentration_model.normed_integrated_concentration(start, time2)
            elif time1 <= start and stop < time2:
                exposure += self.concentration_model.normed_integrated_concentration(start, stop)
        return exposure

    def concentration(self, time: float) -> _VectorisedFloat:
        """
        Virus exposure concentration, as a function of time.

        It considers the long-range concentration with the
        contribution of the short-range concentration.
        """
        concentration = self.concentration_model.concentration(time)
        for interaction in self.short_range:
            concentration += interaction.short_range_concentration(self.concentration_model, time)
        return concentration

    def long_range_deposited_exposure_between_bounds(self, time1: float, time2: float) -> _VectorisedFloat:
        deposited_exposure = 0.

        emission_rate_per_aerosol_per_person = \
            self.concentration_model.infected.emission_rate_per_aerosol_per_person_when_present()
        aerosols = self.concentration_model.infected.aerosols()
        fdep = self.long_range_fraction_deposited()

        diameter = self.concentration_model.infected.particle.diameter

        if not np.isscalar(diameter) and diameter is not None:
            # We compute first the mean of all diameter-dependent quantities
            # to perform properly the Monte-Carlo integration over
            # particle diameters (doing things in another order would
            # lead to wrong results for the probability of infection).
            dep_exposure_integrated = np.array(self._long_range_normed_exposure_between_bounds(time1, time2) *
                                                aerosols *
                                                fdep).mean()
        else:
            # In the case of a single diameter or no diameter defined,
            # one should not take any mean at this stage.
            dep_exposure_integrated = self._long_range_normed_exposure_between_bounds(time1, time2)*aerosols*fdep

        # Then we multiply by the diameter-independent quantity emission_rate_per_aerosol_per_person,
        # and parameters of the vD equation (i.e. BR_k and n_in).
        deposited_exposure += (dep_exposure_integrated *
                emission_rate_per_aerosol_per_person *
                self.exposed.activity.inhalation_rate *
                (1 - self.exposed.mask.inhale_efficiency()))

        return deposited_exposure

    def deposited_exposure_between_bounds(self, time1: float, time2: float) -> _VectorisedFloat:
        """
        The number of virus per m^3 deposited on the respiratory tract
        between any two times.

        Considers a contribution between the short-range and long-range exposures:
        It calculates the deposited exposure given a short-range interaction (if any).
        Then, the deposited exposure given the long-range interactions is added to the
        initial deposited exposure.
        """
        deposited_exposure: _VectorisedFloat = 0.
        for interaction in self.short_range:
            start, stop = interaction.extract_between_bounds(time1, time2)
            short_range_jet_exposure = interaction._normed_jet_exposure_between_bounds(start, stop)
            short_range_lr_exposure = interaction._normed_interpolated_longrange_exposure_between_bounds(
                                        self.concentration_model, start, stop)
            dilution = interaction.dilution_factor()

            fdep = interaction.expiration.particle.fraction_deposited(evaporation_factor=1.0)
            diameter = interaction.expiration.particle.diameter

            # Aerosols not considered given the formula for the initial
            # concentration at mouth/nose.
            if diameter is not None and not np.isscalar(diameter):
                # We compute first the mean of all diameter-dependent quantities
                # to perform properly the Monte-Carlo integration over
                # particle diameters (doing things in another order would
                # lead to wrong results for the probability of infection).
                this_deposited_exposure = (np.array(short_range_jet_exposure
                    * fdep).mean()
                    - np.array(short_range_lr_exposure * fdep).mean()
                    * self.concentration_model.infected.activity.exhalation_rate)
            else:
                # In the case of a single diameter or no diameter defined,
                # one should not take any mean at this stage.
                this_deposited_exposure = (short_range_jet_exposure * fdep
                    - short_range_lr_exposure * fdep
                    * self.concentration_model.infected.activity.exhalation_rate)

            # Multiply by the (diameter-independent) inhalation rate
            deposited_exposure += (this_deposited_exposure *
                                   interaction.activity.inhalation_rate
                                   /dilution)

        # Then we multiply by the emission rate without the BR contribution (and conversion factor),
        # and parameters of the vD equation (i.e. n_in).
        deposited_exposure *= (
            (self.concentration_model.infected.emission_rate_per_aerosol_per_person_when_present() / (
             self.concentration_model.infected.activity.exhalation_rate * 10**6)) *                 
            (1 - self.exposed.mask.inhale_efficiency()))
        # Long-range concentration
        deposited_exposure += self.long_range_deposited_exposure_between_bounds(time1, time2)

        return deposited_exposure

    def _deposited_exposure_list(self):
        """
        The number of virus per m^3 deposited on the respiratory tract.
        """
        population_change_times = self.population_state_change_times()

        deposited_exposure = []
        for start, stop in zip(population_change_times[:-1], population_change_times[1:]):
            deposited_exposure.append(self.deposited_exposure_between_bounds(start, stop))

        return deposited_exposure

    def deposited_exposure(self) -> _VectorisedFloat:
        """
        The number of virus per m^3 deposited on the respiratory tract.
        """
        return np.sum(self._deposited_exposure_list(), axis=0) * self.repeats

    def _infection_probability_list(self):
        # Viral dose (vD)
        vD_list = self._deposited_exposure_list()

        # oneoverln2 multiplied by ID_50 corresponds to ID_63.
        infectious_dose = oneoverln2 * self.concentration_model.virus.infectious_dose

        # Probability of infection.
        return [(1 - np.exp(-((vD * (1 - self.exposed.host_immunity))/(infectious_dose *
                self.concentration_model.virus.transmissibility_factor)))) for vD in vD_list]

    @method_cache
    def infection_probability(self) -> _VectorisedFloat:
        return (1 - np.prod([1 - prob for prob in self._infection_probability_list()], axis = 0)) * 100

    def total_probability_rule(self) -> _VectorisedFloat:
        if (isinstance(self.concentration_model.infected.number, IntPiecewiseConstant) or
                isinstance(self.exposed.number, IntPiecewiseConstant)):
                raise NotImplementedError("Cannot compute total probability "
                        "(including incidence rate) with dynamic occupancy")

        if (self.geographical_data.geographic_population != 0 and self.geographical_data.geographic_cases != 0):
            sum_probability = 0.0

            # Create an equivalent exposure model but changing the number of infected cases.
            total_people = self.concentration_model.infected.number + self.exposed.number
            max_num_infected = (total_people if total_people < 10 else 10)
            # The influence of a higher number of simultainious infected people (> 4 - 5) yields an almost negligible contirbution to the total probability.
            # To be on the safe side, a hard coded limit with a safety margin of 2x was set.
            # Therefore we decided a hard limit of 10 infected people.
            for num_infected in range(1, max_num_infected + 1):
                exposure_model = nested_replace(
                    self, {'concentration_model.infected.number': num_infected}
                )
                prob_ind = exposure_model.infection_probability().mean() / 100
                n = total_people - num_infected
                # By means of the total probability rule
                prob_at_least_one_infected = 1 - (1 - prob_ind)**n
                sum_probability += (prob_at_least_one_infected *
                    self.geographical_data.probability_meet_infected_person(self.concentration_model.infected.virus, num_infected, total_people))
            return sum_probability * 100
        else:
            return 0

    def expected_new_cases(self) -> _VectorisedFloat:
        """
        The expected_new_cases may provide one or two different outputs:
            1) Long-range exposure: take the infection_probability and multiply by the occupants exposed to long-range. 
            2) Short- and long-range exposure: take the infection_probability of long-range multiplied by the occupants exposed to long-range only, 
               plus the infection_probability of short- and long-range multiplied by the occupants exposed to short-range only.

        Currently disabled when dynamic occupancy is defined for the exposed population.
        """

        if (isinstance(self.concentration_model.infected.number, IntPiecewiseConstant) or
            isinstance(self.exposed.number, IntPiecewiseConstant)):
            raise NotImplementedError("Cannot compute expected new cases "
                    "with dynamic occupancy")

        if self.short_range != ():
            new_cases_long_range = nested_replace(self, {'short_range': [],}).infection_probability() * (self.exposed.number - self.exposed_to_short_range)
            return (new_cases_long_range + (self.infection_probability() * self.exposed_to_short_range)) / 100

        return self.infection_probability() * self.exposed.number / 100

    def reproduction_number(self) -> _VectorisedFloat:
        """
        The reproduction number can be thought of as the expected number of
        cases directly generated by one infected case in a population.

        Currently disabled when dynamic occupancy is defined for both the infected and exposed population.
        """

        if (isinstance(self.concentration_model.infected.number, IntPiecewiseConstant) or
            isinstance(self.exposed.number, IntPiecewiseConstant)):
            raise NotImplementedError("Cannot compute reproduction number "
                    "with dynamic occupancy")

        if self.concentration_model.infected.number == 1:
            return self.expected_new_cases()

        # Create an equivalent exposure model but with precisely
        # one infected case.
        single_exposure_model = nested_replace(
            self, {
                'concentration_model.infected.number': 1}
        )

        return single_exposure_model.expected_new_cases()
    


# ------------------ Default form values ----------------------

# Used to declare when an attribute of a class must have a value provided, and
# there should be no default value used.
NO_DEFAULT = object()

#: The default values for undefined fields. Note that the defaults here
#: and the defaults in the html form must not be contradictory.
DEFAULTS = {
    'CO2_fitting_result': '{}',
    'activity_type': 'office',
    'air_changes': 0.,
    'air_supply': 0.,
    'arve_sensors_option': False,
    'ascertainment_bias': 'confidence_low',
    'calculator_version': NO_DEFAULT,
    'ceiling_height': 0.,
    'conditional_probability_viral_loads': False,
    'dynamic_exposed_occupancy': '[]',
    'dynamic_infected_occupancy': '[]',
    'event_month': 'January',
    'exposed_coffee_break_option': 'coffee_break_0',
    'exposed_coffee_duration': 5,
    'exposed_finish': '17:30',
    'exposed_lunch_finish': '13:30',
    'exposed_lunch_option': True,
    'exposed_lunch_start': '12:30',
    'exposed_start': '08:30',
    'exposure_option': 'p_deterministic_exposure',
    'floor_area': 0.,
    'geographic_cases': 0,
    'geographic_population': 0,
    'hepa_amount': 0.,
    'hepa_option': False,
    'humidity': '',
    'infected_coffee_break_option': 'coffee_break_0',
    'infected_coffee_duration': 5,
    'infected_dont_have_breaks_with_exposed': False,
    'infected_finish': '17:30',
    'infected_lunch_finish': '13:30',
    'infected_lunch_option': True,
    'infected_lunch_start': '12:30',
    'infected_people': 1,
    'infected_start': '08:30',
    'inside_temp': NO_DEFAULT,
    'location_latitude': NO_DEFAULT,
    'location_longitude': NO_DEFAULT,
    'location_name': NO_DEFAULT,
    'mask_type': 'Type I',
    'mask_wearing_option': 'mask_off',
    'mechanical_ventilation_type': 'not-applicable',
    'occupancy_format': 'static',
    'opening_distance': 0.,
    'precise_activity': '{}',
    'room_heating_option': False,
    'room_number': NO_DEFAULT,
    'room_volume': 0.,
    'sensor_in_use': '',
    'short_range_interactions': '[]',
    'short_range_occupants': 0,
    'short_range_option': 'short_range_no',
    'simulation_name': NO_DEFAULT,
    'specific_breaks': '{}',
    'total_people': NO_DEFAULT,
    'vaccine_booster_option': False,
    'vaccine_booster_type': 'AZD1222_(AstraZeneca)',
    'vaccine_option': False,
    'vaccine_type': 'AZD1222_(AstraZeneca)',
    'ventilation_type': 'no_ventilation',
    'virus_type': 'SARS_CoV_2',
    'volume_type': NO_DEFAULT,
    'window_height': 0.,
    'window_opening_regime': 'windows_open_permanently',
    'window_type': 'window_sliding',
    'window_width': 0.,
    'windows_duration': 10.,
    'windows_frequency': 60.,
    'windows_number': 0
}

# ------------------ Activities ----------------------

# ------------------ Validation ----------------------
COFFEE_OPTIONS_INT = {'coffee_break_0': 0, 'coffee_break_1': 1,
                      'coffee_break_2': 2, 'coffee_break_4': 4}
CONFIDENCE_LEVEL_OPTIONS = {'confidence_low': 10,
                            'confidence_medium': 5, 'confidence_high': 2}
MECHANICAL_VENTILATION_TYPES = {
    'mech_type_air_changes', 'mech_type_air_supply', 'not-applicable'}
MASK_WEARING_OPTIONS = {'mask_on', 'mask_off'}
MONTH_NAMES = [
    'January', 'February', 'March', 'April', 'May', 'June', 'July',
    'August', 'September', 'October', 'November', 'December',
]
VACCINE_BOOSTER_TYPE = ['AZD1222_(AstraZeneca)', 'Ad26.COV2.S_(Janssen)', 'BNT162b2_(Pfizer)', 'BNT162b2_(Pfizer)_(4th_dose)', 'BNT162b2_(Pfizer)_and_mRNA-1273_(Moderna)',
                        'BNT162b2_(Pfizer)_or_mRNA-1273_(Moderna)', 'BNT162b2_(Pfizer)_or_mRNA-1273_(Moderna)_(4th_dose)', 'CoronaVac_(Sinovac)', 'Coronavac_(Sinovac)', 'Sinopharm',
                        'mRNA-1273_(Moderna)', 'mRNA-1273_(Moderna)_(4th_dose)', 'Other']
VACCINE_TYPE = ['Ad26.COV2.S_(Janssen)', 'Any_mRNA_-_heterologous', 'AZD1222_(AstraZeneca)', 'AZD1222_(AstraZeneca)_and_any_mRNA_-_heterologous', 'AZD1222_(AstraZeneca)_and_BNT162b2_(Pfizer)',
                'BBIBP-CorV_(Beijing_CNBG)', 'BNT162b2_(Pfizer)', 'BNT162b2_(Pfizer)_and_mRNA-1273_(Moderna)', 'CoronaVac_(Sinovac)', 'CoronaVac_(Sinovac)_and_AZD1222_(AstraZeneca)', 'Covishield',
                'mRNA-1273_(Moderna)', 'Sputnik_V_(Gamaleya)', 'CoronaVac_(Sinovac)_and_BNT162b2_(Pfizer)']
VENTILATION_TYPES = {'from_fitting', 'natural_ventilation', 
                     'mechanical_ventilation', 'no_ventilation'}
VOLUME_TYPES = {'room_volume_explicit', 'room_volume_from_dimensions'}
WINDOWS_OPENING_REGIMES = {'windows_open_permanently',
                           'windows_open_periodically', 'not-applicable'}
WINDOWS_TYPES = {'window_sliding', 'window_hinged', 'not-applicable'}



import logging
import typing
import requests

from ..store.data_registry import DataRegistry

logger = logging.getLogger("DATA")


class DataService:
    """Responsible for fetching data from the data service endpoint."""

    # Cached access token
    _access_token: typing.Optional[str] = None

    def __init__(
        self,
        host: str,
    ):
        self._host = host

    @classmethod
    def create(cls, host: str = "https://caimira-data-api-qa.app.cern.ch"):
        """Factory."""
        return cls(host)

    def _fetch(self):

        headers = {
            "Content-Type": "application/json",
        }
        url = f"{self._host}/data"

        try:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
            if response.status_code == 200:
                json_body = response.json()
                logger.debug(f"Data service call: {url}. Response: {json_body}")
                return json_body
            else:
                logger.error(
                    f"Unexpected error when fetching data. Response status code: {response.status_code}, body: f{response.text}"
                )
        except requests.exceptions.RequestException as e:
            logger.exception(e)


    def update_registry(self, registry: DataRegistry):
        data = self._fetch()
        if data:
            registry.update(data["data"], version=data["version"])
        else:
            logger.error("Could not fetch fresh data from the data service.")



# """
# Entry point for the CAiMIRA application
# """

import argparse
import tornado.ioloop
import tornado.web
import tornado.log
import logging
from caimira.api.routes.routes import routes

logging.basicConfig(format="%(message)s", level=logging.INFO)


class Application(tornado.web.Application):
    def __init__(self, debug):
        settings = dict(
            debug=debug,
        )
        super().__init__(routes, **settings)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--no-debug", help="Don't enable debug mode",
        action="store_false",
    )
    parser.add_argument(
        "--port",
        help="The port to listen on",
        default="8081"
    )
    args = parser.parse_args()
    debug = args.no_debug

    app = Application(debug=debug)
    app.listen(args.port)
    logging.info(f"Tornado API server is running on port {args.port}")
    tornado.ioloop.IOLoop.current().start()



import numpy as np
from caimira.calculator.models import models
from .weather import wx_data, nearest_wx_station

MONTH_NAMES = [
    'January', 'February', 'March', 'April', 'May', 'June', 'July',
    'August', 'September', 'October', 'November', 'December',
]


def get_hourly_temperatures_celsius_per_hour(coordinates):
    wx_station_id = nearest_wx_station(
        longitude=coordinates[1], latitude=coordinates[0])[0]
    # Average temperature of each month, hour per hour (from midnight to 11 pm)
    return {month.replace(month, MONTH_NAMES[i][:3]):
            [t - 273.15 for t in temp] for i, (month, temp)
            in enumerate(wx_data()[wx_station_id].items())}


# Load the weather data (temperature in kelvin) for Geneva.
geneva_coordinates = (46.204391, 6.143158)
local_hourly_temperatures_celsius_per_hour = get_hourly_temperatures_celsius_per_hour(
    geneva_coordinates)


# Geneva hourly temperatures as piecewise constant function (in Kelvin).
GenevaTemperatures_hourly = {
    month: models.PiecewiseConstant(
        # NOTE:  It is important that the time type is float, not np.float, in
        # order to allow hashability (for caching).
        tuple(float(time) for time in range(25)),
        tuple(273.15 + np.array(temperatures)),
    )
    for month, temperatures in local_hourly_temperatures_celsius_per_hour.items()
}


# Same Geneva temperatures on a finer temperature mesh (every 6 minutes).
GenevaTemperatures = {
    month: GenevaTemperatures_hourly[month].refine(refine_factor=10)
    for month, temperatures in local_hourly_temperatures_celsius_per_hour.items()
}


# ------- VACCINATION DATA -------

# From data available in Results of COVID-19 Vaccine Effectiveness
# Studies: An Ongoing Systematic Review - Updated September 8, 2022.
# https://view-hub.org/resources
vaccine_primary_host_immunity = {
    'AZD1222_(AstraZeneca)': 0.589293,
    'AZD1222_(AstraZeneca)_and_BNT162b2_(Pfizer)': 0.7865,
    'AZD1222_(AstraZeneca)_and_any_mRNA_-_heterologous': 0.81,
    'Ad26.COV2.S_(Janssen)': 0.551278,
    'Any_mRNA_-_heterologous': 0.93875,
    'BBIBP-CorV_(Beijing_CNBG)': 0.4325,
    'BNT162b2_(Pfizer)': 0.660272,
    'BNT162b2_(Pfizer)_and_mRNA-1273_(Moderna)': 0.586319,
    'CoronaVac_(Sinovac)': 0.317333,
    'CoronaVac_(Sinovac)_and_AZD1222_(AstraZeneca)': 0.472,
    'CoronaVac_(Sinovac)_and_AZD1222_(AstraZeneca)_-_heterologous': 0.74,
    'CoronaVac_(Sinovac)_and_BNT162b2_(Pfizer)': 0.7965,
    'Covishield': 0.98,
    'Sputnik_V_(Gamaleya)': 0.696,
    'mRNA-1273_(Moderna)': 0.730148,
}

vaccine_booster_host_immunity = [
    {'primary series vaccine': 'AZD1222_(AstraZeneca)', 'booster vaccine': 'AZD1222_(AstraZeneca)', 'VE': 0.500429},
    {'primary series vaccine': 'AZD1222_(AstraZeneca)', 'booster vaccine': 'BNT162b2_(Pfizer)', 'VE': 0.537818},
    {'primary series vaccine': 'AZD1222_(AstraZeneca)', 'booster vaccine': 'BNT162b2_(Pfizer)_or_mRNA-1273_(Moderna)', 'VE': 0.284},
    {'primary series vaccine': 'AZD1222_(AstraZeneca)', 'booster vaccine': 'mRNA-1273_(Moderna)', 'VE': 0.709143},
    {'primary series vaccine': 'Ad26.COV2.S_(Janssen)', 'booster vaccine': 'Ad26.COV2.S_(Janssen)', 'VE': 0.492667},
    {'primary series vaccine': 'Ad26.COV2.S_(Janssen)', 'booster vaccine': 'BNT162b2_(Pfizer)_or_mRNA-1273_(Moderna)', 'VE': 0.79},
    {'primary series vaccine': 'BNT162b2_(Pfizer)', 'booster vaccine': 'AZD1222_(AstraZeneca)', 'VE': 0.801},
    {'primary series vaccine': 'BNT162b2_(Pfizer)', 'booster vaccine': 'BNT162b2_(Pfizer)', 'VE': 0.60712},
    {'primary series vaccine': 'BNT162b2_(Pfizer)', 'booster vaccine': 'BNT162b2_(Pfizer)_or_mRNA-1273_(Moderna)', 'VE': 0.632633},
    {'primary series vaccine': 'BNT162b2_(Pfizer)', 'booster vaccine': 'mRNA-1273_(Moderna)', 'VE': 0.716786},
    {'primary series vaccine': 'BNT162b2_(Pfizer)_and_mRNA-1273_(Moderna)', 'booster vaccine': 'BNT162b2_(Pfizer)_and_mRNA-1273_(Moderna)', 'VE': 0.645},
    {'primary series vaccine': 'BNT162b2_(Pfizer)_(3_doses)', 'booster vaccine': 'BNT162b2_(Pfizer)_(4th_dose)', 'VE': 0.962},
    {'primary series vaccine': 'CoronaVac_(Sinovac)', 'booster vaccine': 'AZD1222_(AstraZeneca)', 'VE': 0.9405},
    {'primary series vaccine': 'CoronaVac_(Sinovac)', 'booster vaccine': 'BNT162b2_(Pfizer)', 'VE': 0.690563},
    {'primary series vaccine': 'CoronaVac_(Sinovac)', 'booster vaccine': 'CoronaVac_(Sinovac)', 'VE': 0.52225},
    {'primary series vaccine': 'mRNA-1273_(Moderna)', 'booster vaccine': 'BNT162b2_(Pfizer)', 'VE': 0.842143},
    {'primary series vaccine': 'mRNA-1273_(Moderna)', 'booster vaccine': 'BNT162b2_(Pfizer)_or_mRNA-1273_(Moderna)', 'VE': 0.632633},
    {'primary series vaccine': 'mRNA-1273_(Moderna)', 'booster vaccine': 'mRNA-1273_(Moderna)', 'VE': 0.633238}
]



import typing

import numpy as np
from sklearn.neighbors import KernelDensity

from caimira.calculator.models import models

# Declare a float array type of a given size.
# There is no better way to declare this currently, unfortunately.
float_array_size_n = np.ndarray


class SampleableDistribution:
    def generate_samples(self, size: int) -> float_array_size_n:
        raise NotImplementedError()


class Normal(SampleableDistribution):
    """
    Defines a normal (i.e. Gaussian) distribution
    """
    def __init__(self, mean: float, standard_deviation: float):
        self.mean = mean
        self.standard_deviation = standard_deviation

    def generate_samples(self, size: int) -> float_array_size_n:
        return np.random.normal(self.mean, self.standard_deviation, size=size)


class Uniform(SampleableDistribution):
    """
    Defines a continuous uniform distribution
    """
    def __init__(self, low: float, high: float):
        self.low = low
        self.high = high

    def generate_samples(self, size: int) -> float_array_size_n:
        return np.random.uniform(self.low, self.high, size=size)


class LogNormal(SampleableDistribution):
    """
    Defines a lognormal distribution (i.e. Gaussian distribution vs. the
    natural logarithm of the random variable)
    """

    def __init__(self, mean_gaussian: float, standard_deviation_gaussian: float):
        # these are resp. the mean and std. deviation of the underlying
        # Gaussian distribution
        self.mean_gaussian = mean_gaussian
        self.standard_deviation_gaussian = standard_deviation_gaussian

    def generate_samples(self, size: int) -> float_array_size_n:
        return np.random.lognormal(self.mean_gaussian,
                                   self.standard_deviation_gaussian,
                                   size=size)


class Custom(SampleableDistribution):
    """
    Defines a distribution which follows a custom curve vs. the random
    variable. Uses a simple algorithm. This is appropriate for a smooth
    distribution function.
    Note: in max_function, a value slightly above the maximum of the distribution
    function should be provided.
    """
    def __init__(self, bounds: typing.Tuple[float, float],
                 function: typing.Callable, max_function: float):
        self.bounds = bounds
        self.function = function
        self.max_function = max_function

    def generate_samples(self, size: int) -> float_array_size_n:
        fvalue = np.random.uniform(0,self.max_function,size)
        x = np.random.uniform(*self.bounds,size)
        invalid = np.where(fvalue>self.function(x))[0]
        while len(invalid)>0:
            fvalue[invalid] = np.random.uniform(0,self.max_function,len(invalid))
            x[invalid] = np.random.uniform(*self.bounds,len(invalid))
            invalid = np.where(fvalue>self.function(x))[0]

        return x


class LogCustom(SampleableDistribution):
    """
    Defines a distribution which follows a custom curve vs. the log (in base 10)
    of the random variable. Uses a simple algorithm. This is appropriate for a smooth
    distribution function.
    Note: in max_function, a value slightly above the maximum of the distribution
    function should be provided.
    """
    def __init__(self, bounds: typing.Tuple[float, float],
                 function: typing.Callable, max_function: float):
        self.bounds = bounds
        self.function = function
        self.max_function = max_function

    def generate_samples(self, size: int) -> float_array_size_n:
        fvalue = np.random.uniform(0,self.max_function,size)
        x = np.random.uniform(*self.bounds,size)
        invalid = np.where(fvalue>self.function(x))[0]
        while len(invalid)>0:
            fvalue[invalid] = np.random.uniform(0,self.max_function,len(invalid))
            x[invalid] = np.random.uniform(*self.bounds,len(invalid))
            invalid = np.where(fvalue>self.function(x))[0]

        return 10 ** x


class CustomKernel(SampleableDistribution):
    """
    Defines a distribution which follows a custom curve vs. the
    random variable. Uses a Gaussian kernel density fit. This is more
    appropriate for a noisy distribution function.
    """
    def __init__(self, variable: float_array_size_n,
                 frequencies: float_array_size_n,
                 kernel_bandwidth: float):
        # these are resp. the random variable, the distribution
        # frequencies at these values, and the bandwidth of the Gaussian
        # kernel
        self.variable = variable
        self.frequencies = frequencies
        self.kernel_bandwidth = kernel_bandwidth

    def generate_samples(self, size: int) -> float_array_size_n:
        kde_model = KernelDensity(kernel='gaussian',
                                  bandwidth=self.kernel_bandwidth)
        kde_model.fit(self.variable.reshape(-1, 1),
                      sample_weight=self.frequencies)
        return kde_model.sample(n_samples=size)[:, 0]


class LogCustomKernel(SampleableDistribution):
    """
    Defines a distribution which follows a custom curve vs. the log
    (in base 10) of the random variable. Uses a Gaussian kernel density
    fit. This is more appropriate for a noisy distribution function.
    """
    def __init__(self, log_variable: float_array_size_n,
                 frequencies: float_array_size_n,
                 kernel_bandwidth: float):
        # these are resp. the log of the random variable, the distribution
        # frequencies at these values, and the bandwidth of the Gaussian
        # kernel
        self.log_variable = log_variable
        self.frequencies = frequencies
        self.kernel_bandwidth = kernel_bandwidth

    def generate_samples(self, size: int) -> float_array_size_n:
        kde_model = KernelDensity(kernel='gaussian',
                                  bandwidth=self.kernel_bandwidth)
        kde_model.fit(self.log_variable.reshape(-1, 1),
                      sample_weight=self.frequencies)
        return 10 ** kde_model.sample(n_samples=size)[:, 0]


_VectorisedFloatOrSampleable = typing.Union[
    SampleableDistribution, models._VectorisedFloat,
]



import cProfile
import glob
import json
import logging
import os
import pstats
import shutil
import typing
from datetime import datetime
from enum import Enum
from functools import wraps
from pstats import f8, func_std_string # type: ignore

from pyinstrument import Profiler as PyInstrumentProfiler

logger = logging.getLogger("Profiler")


class Profilers(Enum):
    PYINSTRUMENT = "pyinstrument"
    CPROFILER = "cprofiler"

    @staticmethod
    def from_str(value):
        if value == Profilers.PYINSTRUMENT.value:
            return Profilers.PYINSTRUMENT
        elif value == Profilers.CPROFILER.value:
            return Profilers.CPROFILER
        else:
            raise NotImplementedError


class PyInstrumentWrapper:
    profiler = PyInstrumentProfiler(async_mode=True)

    @property
    def is_running(self):
        return self.profiler.is_running

    def start(self):
        self.profiler.start()

    def stop(self):
        self.profiler.stop()
        report_html = self.profiler.output_html(timeline=True)
        report_id = str(abs(hash(report_html)))
        return report_id, Profilers.PYINSTRUMENT.value, report_html


class CProfilerWrapper:
    profiler = cProfile.Profile()
    _is_running = False

    @property
    def is_running(self):
        return self._is_running

    def start(self):
        self.profiler.enable()
        self._is_running = True

    def stop(self):
        self.profiler.disable()
        self._is_running = False

        ps = pstats.Stats(self.profiler).sort_stats("cumulative")

        report_html = "<p>"
        report_html += f"{ps.total_calls} function calls"
        if ps.total_calls != ps.prim_calls:
            report_html += " (%d primitive calls)" % ps.prim_calls
        report_html += " in %.3f seconds" % ps.total_tt
        report_html += "</p>"

        report_html += "<p>"
        report_html += f"Ordered by: {ps.sort_type}"
        report_html += "</p>"

        report_html += "<table><thead><tr>"
        report_html += "<td>ncalls</td><td>tottime</td><td>percall</td><td>cumtime</td><td>percall</td><td>filename:lineno(function)</td>"
        report_html += "</tr></thead><tbody>"
        for func in ps.stats:
            report_html += "<tr>"
            cc, nc, tt, ct, callers = ps.stats[func]
            c = str(nc)
            if nc != cc:
                c = c + "/" + str(cc)
            report_html += f"<td>{c.rjust(9)}</td>"
            report_html += f"<td>{f8(tt)}</td>"
            report_html += f"<td></td>" if nc == 0 else f"<td>{f8(tt/nc)}</td>"
            report_html += f"<td>{f8(ct)}</td>"
            report_html += f"<td></td>" if cc == 0 else f"<td>{f8(ct/cc)}</td>"
            report_html += f"<td>{func_std_string(func)}</td>"
            report_html += "</tr>"
        report_html += "</tbody></table>"
        report_id = str(abs(hash(report_html)))
        return report_id, Profilers.CPROFILER.value, report_html


class Profiler:
    profiler = None
    profiler_type = None

    def __init__(self, profiler_type):
        self.profiler_type = profiler_type

    def start(self):
        if self.profiler_type == Profilers.PYINSTRUMENT:
            self.profiler = PyInstrumentWrapper()
        elif self.profiler_type == Profilers.CPROFILER:
            self.profiler = CProfilerWrapper()
        else:
            raise NotImplementedError

        self.profiler.start()

    def stop(self, session_name, uri, path="", query="", method=""):
        report_id, report_type, report_html = self.profiler.stop()
        self.profiler = None
        return report_id, dict(
            session_name=session_name,
            ts=datetime.utcnow().strftime("%Y-%m-%d - %H:%M:%S"),
            uri=uri,
            path=path,
            query=query,
            method=method,
            report_id=report_id,
            report_type=report_type,
            report_html=report_html,
        )

    def is_running(self):
        return self.profiler and self.profiler.is_running


class CaimiraProfiler:
    _cache_dirpath: str
    _cache_filepath: str
    CACHE_DIR: str = os.environ.get("CAIMIRA_PROFILER_CACHE_DIR", "/tmp")
    ROOT_URL: str = "/profiler"

    def __init__(self):
        _cache_dir = os.path.join(self.CACHE_DIR, "caimira-profiler")
        self._cache_dirpath = _cache_dir
        self._cache_filepath = os.path.join(_cache_dir, "active")

    def _set_active(self, name: str, profiler_type: Profilers):
        try:
            os.access(self._cache_dirpath, os.W_OK)
        except Exception as e:
            # Handle the exception, e.g., print an error message
            logger.warning(f"cannot write in cache folder {self._cache_dirpath}")

        if not os.path.exists(self._cache_dirpath):
            os.makedirs(self._cache_dirpath)

        with open(self._cache_filepath, "w") as fp:
            json.dump(dict(name=name, profiler_type=profiler_type.value), fp)

    def _set_inactive(self):
        with open(self._cache_filepath, "w") as fp:
            json.dump(dict(), fp)

    @property
    def sessions(self):
        """Return all stored sessions."""
        if not os.path.exists(self._cache_dirpath):
            return {}

        reports_by_session = {}
        json_files = glob.glob(os.path.join(self._cache_dirpath, "*.json"))
        for json_file in json_files:
            with open(json_file) as fp:
                report = json.load(fp)

            session_name = report["session_name"]

            reports_by_session.setdefault(session_name, [])
            reports_by_session[session_name].append(report)

        return reports_by_session

    @property
    def session(self):
        """Return the session name."""
        if not os.path.exists(self._cache_filepath):
            return None

        with open(self._cache_filepath) as fp:
            d = json.load(fp)
            return dict(
                name=d["name"], profiler_type=Profilers.from_str(d["profiler_type"])
            )

    @property
    def is_active(self):
        """Return True if a session is active."""
        enabled = os.environ.get("CAIMIRA_PROFILER_ENABLED", 0)
        if not enabled:
            return False

        if not os.path.exists(self._cache_filepath):
            return False

        try:
            with open(self._cache_filepath) as fp:
                d = json.load(fp)
                return True if d else False
        except json.decoder.JSONDecodeError:
            os.remove(self._cache_filepath)
            return False

    def start_session(
        self,
        name: typing.Optional[str] = None,
        profiler_type: Profilers = Profilers.PYINSTRUMENT,
    ):
        """Start a new session, assigning the given name."""
        now = datetime.utcnow().isoformat()
        name = f"{name}-{now}" if name else now
        self._set_active(name, profiler_type)

    def start_profiler(self):
        """Start the profiler.

        In the context of HTTP requests, this should be called at the beginning of
        each HTTP request.
        """
        if not self.is_active:
            raise Exception("No active session.")

        profiler_type = self.session["profiler_type"]
        profiler = Profiler(profiler_type)
        profiler.start()
        return profiler

    def stop_profiler(
        self,
        profiler,
        uri: str,
        path: typing.Optional[str] = "",
        query: typing.Optional[str] = "",
        method: typing.Optional[str] = "",
    ):
        """Stop the profiler, previously obtained when starting it.

        In the context of HTTP requests, this should be called at the end of
        each HTTP request.
        """
        if not self.is_active:
            raise Exception("No active session.")

        report_id, report = profiler.stop(
            self.session["name"], uri, path, query, method
        )
        with open(os.path.join(self._cache_dirpath, f"{report_id}.json"), "w") as fp:
            json.dump(report, fp)

    def stop_session(self):
        """Stop the current active session."""
        if not self.is_active:
            raise Exception("No active session.")

        self._set_inactive()

    def get_report(self, report_id: str):
        """Return a report by the given id."""
        for reports in self.sessions.values():
            for report in reports:
                if report["report_id"] == report_id:
                    return (
                        Profilers.from_str(report["report_type"]),
                        report["report_html"],
                    )

    def clear_sessions(self):
        """Delete all sessions."""
        if self.is_active:
            raise Exception("Session still active.")

        shutil.rmtree(self._cache_dirpath)


def profile(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        profiler = CaimiraProfiler()
        if profiler.is_active:
            profiler_run = profiler.start_profiler()

        result = func(*args, **kwargs)

        if profiler.is_active:
            profiler.stop_profiler(profiler=profiler_run, uri=func.__name__)

        return result

    return wrapper



import json
import traceback
import sys
from caimira.api.routes.base_handler import BaseRequestHandler
from caimira.api.controller.virus_report_controller import submit_virus_form
from caimira.api.controller.co2_report_controller import request_CO2_transition_times, request_CO2_report


class VirusReportHandler(BaseRequestHandler):
    def post(self):
        try:
            form_data = json.loads(self.request.body)
            arguments = self.request.arguments
            # Report generation parallelism argument
            try:
                report_generation_parallelism = int(arguments['report_generation_parallelism'][0])
            except (ValueError, IndexError, KeyError):
                report_generation_parallelism = None

            report_data = submit_virus_form(form_data, report_generation_parallelism)

            response_data = {
                "status": "success",
                "message": "Results generated successfully",
                "results": report_data,
            }

            self.write(response_data)
        except Exception as e:
            traceback.print_exc()
            self.write_error(status_code=400, exc_info=sys.exc_info())


class CO2SuggestionsHandler(BaseRequestHandler):
    def post(self):
        try:
            form_data = json.loads(self.request.body)
            suggestion_data = request_CO2_transition_times(form_data)

            response_data = {
                "status": "success",
                "message": "Results generated successfully",
                "results": suggestion_data,
            }

            self.write(response_data)
        except Exception as e:
            traceback.print_exc()
            self.write_error(status_code=400, exc_info=sys.exc_info())


class CO2ReportHandler(BaseRequestHandler):
    def post(self):
        try:
            form_data = json.loads(self.request.body)
            report_data = request_CO2_report(form_data)

            response_data = {
                "status": "success",
                "message": "Results generated successfully",
                "results": report_data,
            }

            self.write(response_data)
        except Exception as e:
            traceback.print_exc()
            self.write_error(status_code=400, exc_info=sys.exc_info())



from caimira.calculator.validators.co2.co2_validator import CO2FormData
from caimira.calculator.models.models import CO2DataModel


def build_initial_plot(
        form: CO2FormData,
) -> dict:
    '''
    Initial plot with the suggested ventilation state changes. 
    This method receives the form input and returns the CO2
    plot with the respective transition times.
    '''
    CO2model: CO2DataModel = form.build_model()

    occupancy_transition_times = list(CO2model.occupancy.transition_times)

    ventilation_transition_times: list = form.find_change_points()
    # The entire ventilation changes consider the initial and final occupancy state change
    all_vent_transition_times: list = sorted(
        [occupancy_transition_times[0]] +
        [occupancy_transition_times[-1]] +
        ventilation_transition_times)

    vent_plot_img, vent_plot_data = form.generate_ventilation_plot(
        ventilation_transition_times=ventilation_transition_times,
        occupancy_transition_times=occupancy_transition_times
    )

    context = {
        'transition_times': [round(el, 2) for el in all_vent_transition_times],
        'CO2_plot_img': vent_plot_img,
        'CO2_plot_data': vent_plot_data
    }

    return context


def build_fitting_results(
        form: CO2FormData,
) -> dict:
    '''
    Final fitting results with the respective predictive CO2. 
    This method receives the form input and returns the fitting results
    along with the CO2 plot with the predictive CO2.
    '''
    CO2model: CO2DataModel = form.build_model()

    # Ventilation times after user manipulation from the suggested ventilation state change times.
    ventilation_transition_times = list(CO2model.ventilation_transition_times)

    # The result of the following method is a dict with the results of the fitting
    # algorithm, namely the breathing rate and ACH values. It also returns the
    # predictive CO2 result based on the fitting results.
    context = dict(CO2model.CO2_fit_params())

    vent_plot_img, vent_plot_data = form.generate_ventilation_plot(ventilation_transition_times=ventilation_transition_times[:-1],
                                        predictive_CO2=context['predictive_CO2'])

    # Add the transition times and CO2 plot to the results.
    context['transition_times'] = ventilation_transition_times
    context['CO2_plot_img'] = vent_plot_img
    context['CO2_plot_data'] = vent_plot_data

    return context



from caimira.api.routes.landing_routes import LandingPageHandler
from caimira.api.routes.report_routes import VirusReportHandler, CO2SuggestionsHandler, CO2ReportHandler

routes = [
    (r"/", LandingPageHandler),
    (r"/co2/transition_times", CO2SuggestionsHandler),
    (r"/co2/report", CO2ReportHandler),
    (r"/virus/report", VirusReportHandler),
]



import dataclasses
import html
import logging
import typing
import ast
import json
import re

import numpy as np

from .defaults import DEFAULTS, NO_DEFAULT, COFFEE_OPTIONS_INT
from ..models import models
from ..store.data_registry import DataRegistry

LOG = logging.getLogger(__name__)

minutes_since_midnight = typing.NewType('minutes_since_midnight', int)


@dataclasses.dataclass
class FormData:
    # Static occupancy inputs
    exposed_coffee_break_option: str
    exposed_coffee_duration: int
    exposed_finish: minutes_since_midnight
    exposed_lunch_finish: minutes_since_midnight
    exposed_lunch_option: bool
    exposed_lunch_start: minutes_since_midnight
    exposed_start: minutes_since_midnight
    infected_coffee_break_option: str               #Used if infected_dont_have_breaks_with_exposed
    infected_coffee_duration: int                   #Used if infected_dont_have_breaks_with_exposed
    infected_dont_have_breaks_with_exposed: bool
    infected_finish: minutes_since_midnight
    infected_lunch_finish: minutes_since_midnight   #Used if infected_dont_have_breaks_with_exposed
    infected_lunch_option: bool                     #Used if infected_dont_have_breaks_with_exposed
    infected_lunch_start: minutes_since_midnight    #Used if infected_dont_have_breaks_with_exposed
    infected_start: minutes_since_midnight
    infected_people: int
    occupancy_format: str
    room_volume: float
    specific_breaks: dict
    total_people: int

    # Dynamic occupancy inputs
    dynamic_exposed_occupancy: list
    dynamic_infected_occupancy: list

    data_registry: DataRegistry

    _DEFAULTS: typing.ClassVar[typing.Dict[str, typing.Any]] = DEFAULTS

    @classmethod
    def from_dict(cls, form_data: typing.Dict, data_registry: DataRegistry):
        # Take a copy of the form data so that we can mutate it.
        form_data = form_data.copy()
        form_data.pop('_xsrf', None)
        # Don't let arbitrary unescaped HTML through the net.
        for key, value in form_data.items():
            if isinstance(value, str):
                form_data[key] = html.escape(value)

        for key, default_value in cls._DEFAULTS.items():
            if form_data.get(key, '') == '':
                if default_value is NO_DEFAULT:
                    raise ValueError(f"{key} must be specified")
                form_data[key] = default_value

        for key, value in form_data.items():
            if key in _CAST_RULES_FORM_ARG_TO_NATIVE:
                form_data[key] = _CAST_RULES_FORM_ARG_TO_NATIVE[key](value)

            if key not in cls._DEFAULTS:
                raise ValueError(
                    f'Invalid argument "{html.escape(key)}" given')

        instance = cls(**form_data, data_registry=data_registry)
        instance.validate()
        return instance

    @classmethod
    def to_dict(cls, form: "FormData", strip_defaults: bool = False) -> dict:
        form_dict = {
            field.name: getattr(form, field.name)
            for field in dataclasses.fields(form)
        }

        for attr, value in form_dict.items():
            if attr in _CAST_RULES_NATIVE_TO_FORM_ARG:
                form_dict[attr] = _CAST_RULES_NATIVE_TO_FORM_ARG[attr](value)

        if strip_defaults:
            del form_dict['calculator_version']

            for attr, value in list(form_dict.items()):
                default = cls._DEFAULTS.get(attr, NO_DEFAULT)
                if default is not NO_DEFAULT and value in [default, 'not-applicable']:
                    form_dict.pop(attr)
        return form_dict

    def validate_population_parameters(self):
        # Static occupancy is defined.
        if self.occupancy_format == 'static':        
            # Validate number of infected <= number of total people
            if self.infected_people >= self.total_people:
                raise ValueError(
                    'Number of infected people cannot be greater or equal to the number of total people.')

            # Validate time intervals selected by user
            time_intervals = [
                ['exposed_start', 'exposed_finish'],
                ['infected_start', 'infected_finish'],
            ]
            if self.exposed_lunch_option:
                time_intervals.append(
                    ['exposed_lunch_start', 'exposed_lunch_finish'])
            if self.infected_dont_have_breaks_with_exposed and self.infected_lunch_option:
                time_intervals.append(
                    ['infected_lunch_start', 'infected_lunch_finish'])

            for start_name, end_name in time_intervals:
                start = getattr(self, start_name)
                end = getattr(self, end_name)
                if start > end:
                    raise ValueError(
                        f"{start_name} must be less than {end_name}. Got {start} and {end}.")

            def validate_lunch(start, finish):
                lunch_start = getattr(self, f'{population}_lunch_start')
                lunch_finish = getattr(self, f'{population}_lunch_finish')
                return (start <= lunch_start <= finish and
                        start <= lunch_finish <= finish)

            def get_lunch_mins(population):
                lunch_mins = 0
                if getattr(self, f'{population}_lunch_option'):
                    lunch_mins = getattr(
                        self, f'{population}_lunch_finish') - getattr(self, f'{population}_lunch_start')
                return lunch_mins

            def get_coffee_mins(population):
                coffee_mins = 0
                if getattr(self, f'{population}_coffee_break_option') != 'coffee_break_0':
                    coffee_mins = COFFEE_OPTIONS_INT[getattr(
                        self, f'{population}_coffee_break_option')] * getattr(self, f'{population}_coffee_duration')
                return coffee_mins

            def get_activity_mins(population):
                return getattr(self, f'{population}_finish') - getattr(self, f'{population}_start')

            populations = [
                'exposed', 'infected'] if self.infected_dont_have_breaks_with_exposed else ['exposed']
            for population in populations:
                # Validate lunch time within the activity times.
                if (getattr(self, f'{population}_lunch_option') and
                        not validate_lunch(getattr(self, f'{population}_start'), getattr(
                            self, f'{population}_finish'))
                        ):
                    raise ValueError(
                        f"{population} lunch break must be within presence times."
                    )

                # Length of breaks < length of activity
                if (get_lunch_mins(population) + get_coffee_mins(population)) >= get_activity_mins(population):
                    raise ValueError(
                        f"Length of breaks >= Length of {population} presence."
                    )

                for attr_name, valid_set in [('exposed_coffee_break_option', COFFEE_OPTIONS_INT),
                                            ('infected_coffee_break_option', COFFEE_OPTIONS_INT)]:
                    if getattr(self, attr_name) not in valid_set:
                        raise ValueError(
                            f"{getattr(self, attr_name)} is not a valid value for {attr_name}")
        # Dynamic occupancy is defined.
        elif self.occupancy_format == 'dynamic':
            for dynamic_format in (self.dynamic_infected_occupancy, self.dynamic_exposed_occupancy):
                for occupancy in dynamic_format:
                    # Check if each occupancy entry is a dictionary
                    if not isinstance(occupancy, typing.Dict):
                        raise TypeError(f'Each occupancy entry should be in a dictionary format. Got "{type(occupancy)}".')
                    
                    # Check for required keys in each occupancy entry
                    dict_keys = list(occupancy.keys())
                    if "total_people" not in dict_keys:
                        raise TypeError(f'Unable to fetch "total_people" key. Got "{dict_keys}".')
                    else:
                        value = occupancy["total_people"]
                        # Check if the value is a non-negative integer
                        if not isinstance(value, int):
                            raise ValueError(f'Total number of people should be integer. Got "{type(value)}".')
                        elif not value >= 0:
                            raise ValueError(f'Total number of people should be non-negative. Got "{value}".')
                    
                    if "start_time" not in dict_keys:
                        raise TypeError(f'Unable to fetch "start_time" key. Got "{dict_keys}".')
                    if "finish_time" not in dict_keys:
                        raise TypeError(f'Unable to fetch "finish_time" key. Got "{dict_keys}".')

                    # Validate time format for start_time and finish_time
                    for time_key in ["start_time", "finish_time"]:
                        time = occupancy[time_key]
                        if not re.compile("^(2[0-3]|[01]?[0-9]):([0-5]?[0-9])$").match(time):
                            raise TypeError(f'Wrong time format - "HH:MM". Got "{time}".')
        else:
            raise ValueError(f"'{self.occupancy_format}' is not a valid value for 'self.occupancy_format'. Accepted values are 'static' or 'dynamic'.")

    def validate(self):
        raise NotImplementedError("Subclass must implement")

    def build_model(self, sample_size: typing.Optional[int] = None):
        raise NotImplementedError("Subclass must implement")

    def _compute_breaks_in_interval(self, start, finish, n_breaks, duration) -> models.BoundarySequence_t:
        break_delay = ((finish - start) -
                       (n_breaks * duration)) // (n_breaks+1)
        break_times = []
        end = start
        for n in range(n_breaks):
            begin = end + break_delay
            end = begin + duration
            break_times.append((begin, end))
        return tuple(break_times)

    def exposed_lunch_break_times(self) -> models.BoundarySequence_t:
        result = []
        if self.exposed_lunch_option:
            result.append((self.exposed_lunch_start,
                          self.exposed_lunch_finish))
        return tuple(result)

    def infected_lunch_break_times(self) -> models.BoundarySequence_t:
        if self.infected_dont_have_breaks_with_exposed:
            result = []
            if self.infected_lunch_option:
                result.append((self.infected_lunch_start,
                              self.infected_lunch_finish))
            return tuple(result)
        else:
            return self.exposed_lunch_break_times()

    def exposed_number_of_coffee_breaks(self) -> int:
        return COFFEE_OPTIONS_INT[self.exposed_coffee_break_option]

    def infected_number_of_coffee_breaks(self) -> int:
        return COFFEE_OPTIONS_INT[self.infected_coffee_break_option]

    def _coffee_break_times(self, activity_start, activity_finish, coffee_breaks, coffee_duration, lunch_start, lunch_finish) -> models.BoundarySequence_t:
        time_before_lunch = lunch_start - activity_start
        time_after_lunch = activity_finish - lunch_finish
        before_lunch_frac = time_before_lunch / \
            (time_before_lunch + time_after_lunch)
        n_morning_breaks = round(coffee_breaks * before_lunch_frac)
        breaks = (
            self._compute_breaks_in_interval(
                activity_start, lunch_start, n_morning_breaks, coffee_duration
            )
            + self._compute_breaks_in_interval(
                lunch_finish, activity_finish, coffee_breaks - n_morning_breaks, coffee_duration
            )
        )
        return breaks

    def exposed_coffee_break_times(self) -> models.BoundarySequence_t:
        exposed_coffee_breaks = self.exposed_number_of_coffee_breaks()
        if exposed_coffee_breaks == 0:
            return ()
        if self.exposed_lunch_option:
            breaks = self._coffee_break_times(self.exposed_start, self.exposed_finish, exposed_coffee_breaks,
                                              self.exposed_coffee_duration, self.exposed_lunch_start, self.exposed_lunch_finish)
        else:
            breaks = self._compute_breaks_in_interval(
                self.exposed_start, self.exposed_finish, exposed_coffee_breaks, self.exposed_coffee_duration)
        return breaks

    def infected_coffee_break_times(self) -> models.BoundarySequence_t:
        if self.infected_dont_have_breaks_with_exposed:
            infected_coffee_breaks = self.infected_number_of_coffee_breaks()
            if infected_coffee_breaks == 0:
                return ()
            if self.infected_lunch_option:
                breaks = self._coffee_break_times(self.infected_start, self.infected_finish, infected_coffee_breaks,
                                                  self.infected_coffee_duration, self.infected_lunch_start, self.infected_lunch_finish)
            else:
                breaks = self._compute_breaks_in_interval(
                    self.infected_start, self.infected_finish, infected_coffee_breaks, self.infected_coffee_duration)
            return breaks
        else:
            return self.exposed_coffee_break_times()

    def generate_specific_break_times(self, breaks_dict: dict, target: str) -> models.BoundarySequence_t:
        break_times = []
        for n in breaks_dict[f'{target}_breaks']:
            # Parse break times.
            begin = time_string_to_minutes(n["start_time"])
            end = time_string_to_minutes(n["finish_time"])
            for time in [begin, end]:
                # For a specific break, the infected and exposed presence is the same.
                if not getattr(self, f'{target}_start') < time < getattr(self, f'{target}_finish'):
                    raise ValueError(
                        f'All breaks should be within the simulation time. Got {time_minutes_to_string(time)}.')

            break_times.append((begin, end))
        return tuple(break_times)

    def present_interval(
            self,
            start: int,
            finish: int,
            breaks: typing.Optional[models.BoundarySequence_t] = None,
    ) -> models.Interval:
        """
        Calculate the presence interval given the start and end times (in minutes), and
        a number of monotonic, non-overlapping, but potentially unsorted, breaks (also in minutes).

        """
        if not breaks:
            # If there are no breaks, the interval is the start and end.
            return models.SpecificInterval(((start/60, finish/60),))

        # Order the breaks by their start-time, and ensure that they are monotonic
        # and that the start of one break happens after the end of another.
        break_boundaries: models.BoundarySequence_t = tuple(
            sorted(breaks, key=lambda break_pair: break_pair[0]))

        for break_start, break_end in break_boundaries:
            if break_start >= break_end:
                raise ValueError("Break ends before it begins.")

        prev_break_end = break_boundaries[0][1]
        for break_start, break_end in break_boundaries[1:]:
            if prev_break_end >= break_start:
                raise ValueError(
                    f"A break starts before another ends ({break_start}, {break_end}, {prev_break_end}).")
            prev_break_end = break_end

        present_intervals = []

        current_time = start
        LOG.debug(
            f"starting time march at {_hours2timestring(current_time/60)} to {_hours2timestring(finish/60)}")

        # As we step through the breaks. For each break there are 6 important cases
        # we must cover. Let S=start; E=end; Bs=Break start; Be=Break end:
        #  1. The interval is entirely before the break. S < E <= Bs < Be
        #  2. The interval straddles the start of the break. S < Bs < E <= Be
        #  3. The break is entirely inside the interval. S < Bs < Be <= E
        #  4. The interval is entirely inside the break. Bs <= S < E <= Be
        #  5. The interval straddles the end of the break. Bs <= S < Be <= E
        #  6. The interval is entirely after the break. Bs < Be <= S < E

        for current_break in break_boundaries:
            if current_time >= finish:
                break

            LOG.debug(f"handling break {_hours2timestring(current_break[0]/60)}-{_hours2timestring(current_break[1]/60)} "
                      f" (current time: {_hours2timestring(current_time/60)})")

            break_s, break_e = current_break
            case1 = finish <= break_s
            case2 = current_time < break_s < finish < break_e
            case3 = current_time < break_s < break_e <= finish
            case4 = break_s <= current_time < finish <= break_e
            case5 = break_s <= current_time < break_e < finish
            case6 = break_e <= current_time

            if case1:
                LOG.debug(f"case 1: interval entirely before break")
                present_intervals.append((current_time / 60, finish / 60))
                LOG.debug(f" + added interval {_hours2timestring(present_intervals[-1][0])} "
                          f"- {_hours2timestring(present_intervals[-1][1])}")
                current_time = finish
            elif case2:
                LOG.debug(f"case 2: interval straddles start of break")
                present_intervals.append((current_time / 60, break_s / 60))
                LOG.debug(f" + added interval {_hours2timestring(present_intervals[-1][0])} "
                          f"- {_hours2timestring(present_intervals[-1][1])}")
                current_time = break_e
            elif case3:
                LOG.debug(f"case 3: break entirely inside interval")
                # We add the bit before the break, but not the bit afterwards,
                # as it may hit another break.
                present_intervals.append((current_time / 60, break_s / 60))
                LOG.debug(f" + added interval {_hours2timestring(present_intervals[-1][0])} "
                          f"- {_hours2timestring(present_intervals[-1][1])}")
                current_time = break_e
            elif case4:
                LOG.debug(f"case 4: interval entirely inside break")
                current_time = finish
            elif case5:
                LOG.debug(f"case 5: interval straddles end of break")
                current_time = break_e
            elif case6:
                LOG.debug(f"case 6: interval entirely after the break")

        if current_time < finish:
            LOG.debug("trailing interval")
            present_intervals.append((current_time / 60, finish / 60))
        return models.SpecificInterval(tuple(present_intervals))

    def infected_present_interval(self) -> models.Interval:
        if self.specific_breaks != {}:  # It means the breaks are specific and not predefined
            breaks = self.generate_specific_break_times(
                breaks_dict=self.specific_breaks, target='exposed')
        else:
            breaks = self.infected_lunch_break_times() + self.infected_coffee_break_times()
        return self.present_interval(
            self.infected_start, self.infected_finish,
            breaks=breaks,
        )

    def exposed_present_interval(self) -> models.Interval:
        if self.specific_breaks != {}:  # It means the breaks are specific and not predefined
            breaks = self.generate_specific_break_times(
                breaks_dict=self.specific_breaks, target='exposed')
        else:
            breaks = self.exposed_lunch_break_times() + self.exposed_coffee_break_times()
        return self.present_interval(
            self.exposed_start, self.exposed_finish,
            breaks=breaks,
        )
    
    def generate_dynamic_occupancy(self, dynamic_occupancy: typing.List[typing.Dict[str, typing.Any]]):
        transition_times = []
        values = []
        for occupancy in dynamic_occupancy:
            start_time = time_string_to_minutes(occupancy['start_time'])/60
            finish_time = time_string_to_minutes(occupancy['finish_time'])/60
            transition_times.extend([start_time, finish_time])
            values.append(occupancy['total_people'])

        unique_transition_times_sorted = np.array(sorted(set(transition_times)))

        if len(values) != len(unique_transition_times_sorted) - 1:
            raise ValueError("Cannot compute dynamic occupancy with the provided inputs.")
        
        population_occupancy: models.IntPiecewiseConstant = models.IntPiecewiseConstant(
            transition_times=tuple(unique_transition_times_sorted),
            values=tuple(values)
        )
        return population_occupancy


def _hours2timestring(hours: float):
    # Convert times like 14.5 to strings, like "14:30"
    return f"{int(np.floor(hours)):02d}:{int(np.round((hours % 1) * 60)):02d}"


def time_string_to_minutes(time: str) -> minutes_since_midnight:
    """
    Converts time from string-format to an integer number of minutes after 00:00
    :param time: A string of the form "HH:MM" representing a time of day
    :return: The number of minutes between 'time' and 00:00
    """
    return minutes_since_midnight(60 * int(time[:2]) + int(time[3:]))


def time_minutes_to_string(time: int) -> str:
    """
    Converts time from an integer number of minutes after 00:00 to string-format
    :param time: The number of minutes between 'time' and 00:00
    :return: A string of the form "HH:MM" representing a time of day
    """
    return "{0:0=2d}".format(int(time/60)) + ":" + "{0:0=2d}".format(time % 60)


def string_to_list(s: str) -> list:
    return list(ast.literal_eval(s.replace("&quot;", "\"")))


def list_to_string(l: list) -> str:
    return json.dumps(l)


def string_to_dict(s: str) -> dict:
    return dict(ast.literal_eval(s.replace("&quot;", "\"")))


def dict_to_string(d: dict) -> str:
    return json.dumps(d)


def _safe_int_cast(value) -> int:
    if isinstance(value, int):
        return value
    elif isinstance(value, float) and int(value) == value:
        return int(value)
    elif isinstance(value, str) and value.isdecimal():
        return int(value)
    else:
        raise TypeError(f"Unable to safely cast {value} ({type(value)} type) to int")
    

def _safe_optional_int_cast(value) -> typing.Optional[int]:
    if value is None or value == '':
        return None
    return _safe_int_cast(value)


#: Mapping of field name to a callable which can convert values from form
#: input (URL encoded arguments / string) into the correct type.
_CAST_RULES_FORM_ARG_TO_NATIVE: typing.Dict[str, typing.Callable] = {}

#: Mapping of field name to callable which can convert native type to values
#: that can be encoded to URL arguments.
_CAST_RULES_NATIVE_TO_FORM_ARG: typing.Dict[str, typing.Callable] = {}


def cast_class_fields(cls):
    for _field in dataclasses.fields(cls):
        if _field.type is minutes_since_midnight:
            _CAST_RULES_FORM_ARG_TO_NATIVE[_field.name] = time_string_to_minutes
            _CAST_RULES_NATIVE_TO_FORM_ARG[_field.name] = time_minutes_to_string
        elif _field.type is int:
            _CAST_RULES_FORM_ARG_TO_NATIVE[_field.name] = _safe_int_cast
        elif _field.type is typing.Optional[int]:
            _CAST_RULES_FORM_ARG_TO_NATIVE[_field.name] = _safe_optional_int_cast
        elif _field.type is float:
            _CAST_RULES_FORM_ARG_TO_NATIVE[_field.name] = float
        elif _field.type is bool:
            _CAST_RULES_FORM_ARG_TO_NATIVE[_field.name] = lambda v: v == '1'
            _CAST_RULES_NATIVE_TO_FORM_ARG[_field.name] = int
        elif _field.type is list:
            _CAST_RULES_FORM_ARG_TO_NATIVE[_field.name] = string_to_list
            _CAST_RULES_NATIVE_TO_FORM_ARG[_field.name] = list_to_string
        elif _field.type is dict:
            _CAST_RULES_FORM_ARG_TO_NATIVE[_field.name] = string_to_dict
            _CAST_RULES_NATIVE_TO_FORM_ARG[_field.name] = dict_to_string


cast_class_fields(FormData)



import copy
import dataclasses
import sys
import typing

from caimira.calculator.models import models

from .sampleable import SampleableDistribution, _VectorisedFloatOrSampleable

_ModelType = typing.TypeVar('_ModelType')
dataclass_instance = typing.Any

class MCModelBase(typing.Generic[_ModelType]):
    """
    A model base class for monte carlo types.

    This base class is essentially a declarative description of a caimira.models
    model with a :meth:`.build_model` method to generate an appropriate
    ``caimira.models` model instance on demand.

    """
    _base_cls: typing.Type[dataclass_instance]

    @classmethod
    def _to_vectorized_form(cls, item, size):
        if isinstance(item, SampleableDistribution):
            return item.generate_samples(size)
        elif isinstance(item, MCModelBase):
            # Recurse into other MCModelBase instances by calling their
            # build_model method.
            return item.build_model(size)
        elif isinstance(item, tuple):
            return tuple(cls._to_vectorized_form(sub, size) for sub in item)
        else:
            return item

    def build_model(self, size: int) -> _ModelType:
        """
        Turn this MCModelBase subclass into a caimira.model Model instance
        from which you can then run the model.

        """
        kwargs = {}
        for field in dataclasses.fields(self._base_cls):
            attr = getattr(self, field.name)
            kwargs[field.name] = self._to_vectorized_form(attr, size)
        return self._base_cls(**kwargs)


def _build_mc_model(model: dataclass_instance) -> typing.Type[MCModelBase[_ModelType]]:
    """
    Generate a new MCModelBase subclass for the given caimira.models model.

    """
    fields = []
    for field in dataclasses.fields(model):
        # Note: deepcopy not needed here as we aren't mutating entities beyond
        # the top level.
        new_field = copy.copy(field)
        if field.type is models._VectorisedFloat:  # noqa
            new_field.type = _VectorisedFloatOrSampleable   # type: ignore

        field_type: typing.Any = new_field.type

        if getattr(field_type, '__origin__', None) in [typing.Union, typing.Tuple]:
            # It is challenging to generalise this code, so we provide specific transformations,
            # and raise for unforseen cases.
            if new_field.type == typing.Tuple[models._VentilationBase, ...]:
                VB = getattr(sys.modules[__name__], "_VentilationBase")
                field_type = typing.Tuple[typing.Union[models._VentilationBase, VB], ...]
            elif new_field.type == typing.Tuple[models._ExpirationBase, ...]:
                EB = getattr(sys.modules[__name__], "_ExpirationBase")
                field_type = typing.Tuple[typing.Union[models._ExpirationBase, EB], ...]
            elif new_field.type == typing.Tuple[models.SpecificInterval, ...]:
                SI = getattr(sys.modules[__name__], "SpecificInterval")
                field_type = typing.Tuple[typing.Union[models.SpecificInterval, SI], ...]

            elif new_field.type == typing.Union[int, models.IntPiecewiseConstant]:
                IPC = getattr(sys.modules[__name__], "IntPiecewiseConstant")
                field_type = typing.Union[int, models.IntPiecewiseConstant, IPC]
            elif new_field.type == typing.Union[models.Interval, None]:
                I = getattr(sys.modules[__name__], "Interval")
                field_type = typing.Union[None, models.Interval, I]

            else:
                # Check that we don't need to do anything with this type.
                if hasattr(new_field.type, "__args__"):
                    for item in new_field.type.__args__:
                        if getattr(item, '__module__', None) == 'source.models.models':
                            raise ValueError(
                                f"unsupported type annotation transformation required for {new_field.type}")
        elif field_type.__module__ == 'source.models.models':
            if isinstance(new_field.type, type):
                mc_model = getattr(sys.modules[__name__], new_field.type.__name__)
                field_type = typing.Union[new_field.type, mc_model]
            else:
                raise ValueError(f"Expected a class/type but got {new_field.type}")
        fields.append((new_field.name, field_type, new_field))

    bases = []
    # Update the inheritance/based to use the new MC classes, rather than the caimira.models ones.
    for model_base in model.__bases__:  # type: ignore
        if model_base is object:
            bases.append(MCModelBase)
        else:
            mc_model = getattr(sys.modules[__name__], model_base.__name__)
            bases.append(mc_model)

    cls = dataclasses.make_dataclass(
        model.__name__,  # type: ignore
        fields,  # type: ignore
        bases=tuple(bases),  # type: ignore
        namespace={'_base_cls': model},
        # This thing can be mutable - the calculations live on
        # the wrapped class, not on the MCModelBase.
        frozen=False,
    )
    # Update the module of the generated class to be this one. Without this the
    # module will be "types".
    cls.__module__ = __name__
    return cls


_MODEL_CLASSES = [
    cls for cls in vars(models).values()
    if dataclasses.is_dataclass(cls) and isinstance(cls, type)
]


# Inject the runtime generated MC types into this module.
for _model in _MODEL_CLASSES:
    setattr(sys.modules[__name__], _model.__name__, _build_mc_model(_model)) # type: ignore


# Make sure that each of the models is imported if you do a ``import *``.
__all__ = [_model.__name__ for _model in _MODEL_CLASSES] + ["MCModelBase"] # type: ignore



import concurrent.futures
import functools
import typing

from caimira.calculator.validators.virus.virus_validator import VirusFormData
from caimira.calculator.store.data_registry import DataRegistry
import caimira.calculator.report.virus_report_data as rg


def generate_form_obj(form_data: typing.Dict, data_registry: DataRegistry) -> VirusFormData:
    return VirusFormData.from_dict(
        form_data=form_data,
        data_registry=data_registry,
    )


def generate_report(form_obj: VirusFormData, report_generation_parallelism: typing.Optional[int]) -> typing.Dict:
    return rg.calculate_report_data(
        form=form_obj,
        executor_factory=functools.partial(
            concurrent.futures.ThreadPoolExecutor,
            report_generation_parallelism,
        ),
    )


def submit_virus_form(form_data: typing.Dict, report_generation_parallelism: typing.Optional[int]) -> typing.Dict:
    data_registry: DataRegistry = DataRegistry()

    form_obj: VirusFormData = generate_form_obj(form_data=form_data, data_registry=data_registry)
    report_data: typing.Dict = generate_report(form_obj=form_obj, report_generation_parallelism=report_generation_parallelism)

    # Handle model representation
    if report_data['model']: report_data['model'] = repr(report_data['model'])

    return report_data



import pandas as pd
from tabulate import tabulate

'''
Script file to generate the vaccine effectiveness values.
To generate the primary vaccine effectiveness values, uncomment lines 16-21.
To generate the booster effectiveness values, uncomment lines 26-56.
'''

# Data from 08 Sep. 2022
file_loc = "https://caimira-resources.web.cern.ch/WeeklySummary_COVID19_VE_Studies_08Sep2022_adapted.xlsx"


# ------- PRIMARY VACCINATION ------ #

# df = pd.read_excel(file_loc, sheet_name="Primary_filtered", usecols="A, B, E")

# calculate the VE value
# df = df.drop(df[df['VE'] < 0].index)
# ve_data = df.groupby(['vaccine'])['VE'].mean().divide(100).reset_index()
# print(tabulate(ve_data, headers='keys', tablefmt='psql'))


# ------- BOOSTER VACCINATION ------ #

# df = pd.read_excel(file_loc, sheet_name="Booster_filtered", usecols="A, B, C, F")

# # create df without the '  or  ' substring in primary vaccines
# rows_with_or = df[df['primary series vaccine'].str.contains(' or ')]
# rows_indexes = list(rows_with_or.index)
# df_without_or = df.drop(labels=rows_indexes, axis=0)

# # copy of all the rows that contain '  or  '
# new_rows_with_or = rows_with_or.reset_index().copy()

# # create new dataframe empty
# rows_to_add = pd.DataFrame(columns=rows_with_or.columns)

# # duplicate each row and add it into the new dataframe
# for index, row in new_rows_with_or.iterrows():
#     new_rows_with_or.at[index, 'primary series vaccine'] = row['primary series vaccine'].split(' or ')[0]
#     rows_to_add.loc[index] = new_rows_with_or.loc[index]
#     new_rows_with_or.at[index, 'primary series vaccine'] = row['primary series vaccine'].split(' or ')[1]
#     rows_to_add.loc[len(rows_indexes)+index] = new_rows_with_or.loc[index]

# # merge the dataframe without the '  or  ' with the new dataframe that has the rows divided in two
# final_df = pd.concat([df_without_or, rows_to_add]).reset_index().drop(columns=['index'])

# # calculate the VE value
# final_df = final_df.drop(final_df[final_df['VE'] < 0].index)

# ve_data = final_df.groupby(['primary series vaccine', 'booster vaccine'])['VE'].mean().divide(100).reset_index()

# result = ve_data.to_dict('records')

# print(tabulate(ve_data, headers='keys', tablefmt='psql'))



import dataclasses
import typing


def nested_replace(obj, new_values: typing.Dict[str, typing.Any]):
    """Replace an attribute on a dataclass, much like dataclasses.replace,
    except it supports nested replacement definitions. For example:

    >>> new_obj = nested_replace(obj, {'attr1.sub_attr2.sub_sub_attr3': 4})
    >>> new_obj.attr1.sub_attr2.sub_sub_attr3
    4

    """
    new_inst = obj
    for name, value in new_values.items():
        if '.' in name:
            # Recurse into the desired name and come out with a top-level
            # dataclass which has been updated appropriately.
            name, remainder = name.split('.', 1)
            value = nested_replace(
                getattr(new_inst, name),
                {remainder: value}
            )
        # We have a plain old name. So set it.
        new_inst = dataclasses.replace(new_inst, **{name: value})
    return new_inst


def nested_getattr(obj, name: str):
    """Get an attribute on a dataclass, much like getattr,
    except it supports nested attributes definitions. For example:

    >>> nested_getattr(obj, 'attr1.sub_attr2.sub_sub_attr3')
 
    """
    if '.' in name:
        # Recurse into the desired name and come out with a top-level
        # dataclass which has been updated appropriately.
        name, remainder = name.split('.', 1)
        return nested_getattr(getattr(obj,name), remainder)
    else:
        return getattr(obj,name)


def replace(obj, **changes):
    """
    A version of dataclasses.replace that handles ClassVar declarations.

    See https://bugs.python.org/issue33796.

    """

    orig = obj.__dataclass_fields__
    object.__setattr__(
        obj, '__dataclass_fields__',
        {name: field for name, field in orig.items()
         if field._field_type is not dataclasses._FIELD_CLASSVAR}
    )
    new = dataclasses.replace(obj, **changes)
    object.__setattr__(obj, '__dataclass_fields__', orig)
    return new


def walk_dataclass(model, name=""):
    """
    Recursively walk a dataclass instance, generating (name, obj) pairs for
    attributes and decending into nested dataclasses.

    >>> list(walk_dataclass(obj), 'my_obj')
    [('my_obj.attr_a', <dataclass instance>), ('my_obj.attr_a.sub_attr', <dataclass instance>)]

    """
    if name:
        name = name + '.'
    if not dataclasses.is_dataclass(model):
        raise TypeError(f'Not a dataclass based model: {type(model)}')
    for field in dataclasses.fields(model):
        obj = getattr(model, field.name)
        fq_name = f'{name}{field.name}'
        yield fq_name, obj
        if dataclasses.is_dataclass(obj):
            yield from walk_dataclass(obj, name=fq_name)
